from collections import UserString

from .._type_converter import infer_type, type2str
from ..helpers import InpSectionGeneric
from ..section_labels import TITLE, OPTIONS, REPORT, EVAPORATION, TEMPERATURE, MAP, FILES, ADJUSTMENTS, BACKDROP


def line_iter(lines):
    if isinstance(lines, str):
        lines = lines.split('\n')

    for line in lines:
        line = line.split(';')[0]
        line = line.strip()
        if line == '':  # ignore empty and comment lines
            continue
        else:
            yield line.split()


class TitleSection(UserString, InpSectionGeneric):
    """
    abstract class for ``.inp``-file sections without objects

    :term:`dict-like <mapping>`"
    """
    _section_label = TITLE
    """str: label of the section"""

    def __init__(self, *args, **kwargs):
        UserString.__init__(self, *args, **kwargs)
        self._inp = None

    def set_parent_inp(self, inp):
        self._inp = inp

    def get_parent_inp(self):
        return self._inp

    @classmethod
    def from_inp_lines(cls, lines):
        """
        read ``.inp``-file lines and create an section object

        Args:
            lines (str | list[list[str]]): lines in the section of the ``.inp``-file

        Returns:
            InpSectionGeneric: object of the section
        """
        return cls(lines)

    def to_inp_lines(self, fast=False, sort_objects_alphabetical=False):
        """
        write ``.inp``-file lines of the section object

        Args:
            fast (bool): speeding up conversion

                - :obj:`True`: if no special formation of the input file is needed
                - :obj:`False`: section is converted into a table to prettify string output (slower)

        Returns:
            str: ``.inp``-file lines of the section object
        """
        return str(self)

    def copy(self):
        return self


class OptionSection(InpSectionGeneric):
    """
    Section: [**OPTIONS**]

    Purpose:
        Provides values for various analysis options.

    Format:
        ::

            FLOW_UNITS           CFS*/GPM/MGD/CMS/LPS/MLD
            INFILTRATION         HORTON* / MODIFIED_HORTON / GREEN_AMPT / MODIFIED_GREEN_AMPT / CURVE_NUMBER
            FLOW_ROUTING         STEADY / KINWAVE* / DYNWAVE
            LINK_OFFSETS         DEPTH* / ELEVATION
            FORCE_MAIN_EQUATION  H-W* / D-W
            IGNORE_RAINFALL      YES / NO*
            IGNORE_SNOWMELT      YES / NO*
            IGNORE_GROUNDWATER   YES / NO*
            IGNORE_RDII          YES / NO*
            IGNORE_ROUTING       YES / NO*
            IGNORE_QUALITY       YES / NO*
            ALLOW_PONDING        YES / NO*
            SKIP_STEADY_STATE    YES / NO*
            SYS_FLOW_TOL         value (5)
            LAT_FLOW_TOL         value (5)
            START_DATE           month/day/year (1/1/2002)
            START_TIME           hours:minutes (0:00:00)
            END_DATE             month/day/year (START_DATE)
            END_TIME             hours:minutes (24:00:00)
            REPORT_START_DATE    month/day/year (START_DATE)
            REPORT_START_TIME    hours:minutes (START_TIME)
            SWEEP_START          month/day (1/1)
            SWEEP_END            month/day (12/31)
            DRY_DAYS             days (0)
            REPORT_STEP          hours:minutes:seconds (0:15:00)
            WET_STEP             hours:minutes:seconds (0:05:00)
            DRY_STEP             hours:minutes:seconds (1:00:00)
            ROUTING_STEP         seconds (600)
            LENGTHENING_STEP     seconds (0)
            VARIABLE_STEP        value (0)
            MINIMUM_STEP         seconds (0.5)
            INERTIAL_DAMPING     NONE / PARTIAL / FULL
            NORMAL_FLOW_LIMITED  SLOPE / FROUDE / BOTH*

            MIN_SURFAREA        value (12.566 ft2 (i.e., the area of a 4-ft diameter manhole))
            MIN_SLOPE           value (0)
            MAX_TRIALS          value (8)
            HEAD_TOLERANCE      value (0.0015)
            THREADS             value (1)
            TEMPDIR             directory

        `* .. defaults`

        FLOW_UNITS makes a choice of flow units. Selecting a US flow unit means that all
        other quantities will be expressed in US units, while choosing a metric flow unit will
        force all quantities to be expressed in metric units. The default is CFS.

        INFILTRATION selects a model for computing infiltration of rainfall into the upper
        soil zone of subcatchments. The default model is HORTON.

        FLOW_ROUTING determines which method is used to route flows through the
        drainage system. STEADY refers to sequential steady state routing (i.e. hydrograph
        translation), KINWAVE to kinematic wave routing, DYNWAVE to dynamic wave routing.
        The default routing method is KINWAVE.

        LINK_OFFSETS determines the convention used to specify the position of a link
        offset above the invert of its connecting node. DEPTH indicates that offsets are
        expressed as the distance between the node invert and the link while ELEVATION
        indicates that the absolute elevation of the offset is used. The default is DEPTH.

        FORCE_MAIN_EQUATION establishes whether the Hazen-Williams (H-W) or the
        Darcy-Weisbach (D-W) equation will be used to compute friction losses for
        pressurized flow in conduits that have been assigned a Circular Force Main cross-
        section shape. The default is H-W.

        IGNORE_RAINFALL is set to YES if all rainfall data and runoff calculations should be
        ignored. In this case SWMM only performs flow and pollutant routing based on user-
        supplied direct and dry weather inflows. The default is NO.

        IGNORE_SNOWMELT is set to YES if snowmelt calculations should be ignored when a
        project file contains snow pack objects. The default is NO.

        IGNORE_GROUNDWATER is set to YES if groundwater calculations should be ignored
        when a project file contains aquifer objects. The default is NO.

        IGNORE_RDII is set to YES if rainfall dependent inflow/infiltration should be ignored
        when RDII unit hydrographs and RDII inflows have been supplied to a project file.
        The default is NO.

        IGNORE_ROUTING is set to YES if only runoff should be computed even if the project
        contains drainage system links and nodes. The default is NO.

        IGNORE_QUALITY is set to YES if pollutant washoff, routing, and treatment should be
        ignored in a project that has pollutants defined. The default is NO.

        ALLOW_PONDING determines whether excess water is allowed to collect atop nodes
        and be re-introduced into the system as conditions permit. The default is NO ponding.

        In order for ponding to actually occur at a particular node, a non-zero value for its
        Ponded Area attribute must be used.

        SKIP_STEADY_STATE should be set to YES if flow routing computations should be
        skipped during steady state periods of a simulation during which the last set of
        computed flows will be used. A time step is considered to be in steady state if the
        percent difference between total system inflow and total system outflow is below the
        SYS_FLOW_TOL and the percent difference between current and previous lateral
        inflows are below the LAT_FLOW_TOL. The default for this option is NO.

        SYS_FLOW_TOL is the maximum percent difference between total system inflow and
        total system outflow which can occur in order for the SKIP_STEADY_STATE option to
        take effect. The default is 5 percent.

        LAT_FLOW_TOL is the maximum percent difference between the current and
        previous lateral inflow at all nodes in the conveyance system in order for the
        SKIP_STEADY_STATE option to take effect. The default is 5 percent.

        START_DATE is the date when the simulation begins. If not supplied, a date of
        1/1/2002 is used.

        START_TIME is the time of day on the starting date when the simulation begins. The
        default is 12 midnight (0:00:00).

        END_DATE is the date when the simulation is to end. The default is the start date.
        END_TIME is the time of day on the ending date when the simulation will end. The
        default is 24:00:00.

        REPORT_START_DATE is the date when reporting of results is to begin. The default is
        the simulation start date.

        REPORT_START_TIME is the time of day on the report starting date when reporting is
        to begin. The default is the simulation start time of day.

        SWEEP_START is the day of the year (month/day) when street sweeping operations
        begin. The default is 1/1.

        SWEEP_END is the day of the year (month/day) when street sweeping operations end.
        The default is 12/31.

        DRY_DAYS is the number of days with no rainfall prior to the start of the simulation.
        The default is 0.

        REPORT_STEP is the time interval for reporting of computed results. The default is
        0:15:00.

        WET_STEP is the time step length used to compute runoff from subcatchments
        during periods of rainfall or when ponded water still remains on the surface. The
        default is 0:05:00.

        DRY_STEP is the time step length used for runoff computations (consisting essentially
        of pollutant buildup) during periods when there is no rainfall and no ponded water.
        The default is 1:00:00.

        ROUTING_STEP is the time step length in seconds used for routing flows and water
        quality constituents through the conveyance system. The default is 600 sec (5
        minutes) which should be reduced if using dynamic wave routing. Fractional values
        (e.g., 2.5) are permissible as are values entered in hours:minutes:seconds format.

        LENGTHENING_STEP is a time step, in seconds, used to lengthen conduits under
        dynamic wave routing, so that they meet the Courant stability criterion under full-flow
        conditions (i.e., the travel time of a wave will not be smaller than the specified conduit
        lengthening time step). As this value is decreased, fewer conduits will require
        lengthening. A value of 0 (the default) means that no conduits will be lengthened.

        VARIABLE_STEP is a safety factor applied to a variable time step computed for
        each time period under dynamic wave flow routing. The variable time step is
        computed so as to satisfy the Courant stability criterion for each conduit and yet not
        exceed the ROUTING_STEP value. If the safety factor is 0 (the default), then no
        variable time step is used.

        MINIMUM_STEP is the smallest time step allowed when variable time steps are used
        for dynamic wave flow routing. The default value is 0.5 seconds.

        INERTIAL_DAMPING indicates how the inertial terms in the Saint Venant
        momentum equation will be handled under dynamic wave flow routing. Choosing
        NONE maintains these terms at their full value under all conditions. Selecting
        PARTIAL will reduce the terms as flow comes closer to being critical (and ignores
        them when flow is supercritical). Choosing FULL will drop the terms altogether.

        NORMAL_FLOW_LIMITED specifies which condition is checked to determine if flow in
        a conduit is supercritical and should thus be limited to the normal flow. Use SLOPE to
        check if the water surface slope is greater than the conduit slope, FROUDE to check if
        the Froude number is greater than 1.0, or BOTH to check both conditions. The default
        is BOTH.

        MIN_SURFAREA is a minimum surface area used at nodes when computing changes
        in water depth under dynamic wave routing. If 0 is entered, then the default value of
        12.566 ft2 (i.e., the area of a 4-ft diameter manhole) is used.

        MIN_SLOPE is the minimum value allowed for a conduit’s slope (%). If zero (the
        default) then no minimum is imposed (although SWMM uses a lower limit on
        elevation drop of 0.001 ft (0.00035 m) when computing a conduit slope).

        MAX_TRIALS is the maximum number of trials allowed during a time step to reach
        convergence when updating hydraulic heads at the conveyance system’s nodes. The
        default value is 8.

        HEAD_TOLERANCE is the difference in computed head at each node between
        successive trials below which the flow solution for the current time step is assumed to
        have converged. The default tolerance is 0.005 ft (0.0015 m).

        THREADS is the number of parallel computing threads to use for dynamic wave flow
        routing on machines equipped with multi-core processors. The default is 1.

        TEMPDIR provides the name of a file directory (or folder) where SWMM writes its
        temporary files. If the directory name contains spaces then it should be placed within
        double quotes. If no directory is specified, then the temporary files are written to the
        current directory that the user is working in.

    Args:
        lines (list): section lines from input file

    Returns:
        dict: options
    """
    _label = OPTIONS

    @classmethod
    def from_inp_lines(cls, lines):
        data = cls()
        for key, *line in line_iter(lines):
            key = key.upper()
            assert len(line) == 1
            data[key] = infer_type(line[0])
        return data


class ReportSection(InpSectionGeneric):
    """
    Section: [**REPORT**]

    Purpose:
        Describes the contents of the report file that is produced.

    Formats:
        ::

            INPUT          YES / NO*
            CONTINUITY     YES* / NO
            FLOWSTATS      YES* / NO
            CONTROLS       YES / NO*
            SUBCATCHMENTS  ALL / NONE* / <list of subcatchment names>
            NODES          ALL / NONE* / <list of node names>
            LINKS          ALL / NONE* / <list of link names>
            LID            Name Subcatch Fname

        `* .. defaults`

    Remarks:
        INPUT
            specifies whether or not a summary of the input data should be provided in the output report.
            The default is NO.
        CONTINUITY
            specifies whether continuity checks should be reported or not. The default is YES.
        FLOWSTATS
            specifies whether summary flow statistics should be reported or not. The default is YES.
        CONTROLS
            specifies whether all control actions taken during a simulation should be listed or not. The default is NO.
        SUBCATCHMENTS
            gives a list of subcatchments whose results are to be reported. The default is NONE.
        NODES
            gives a list of nodes whose results are to be reported. The default is NONE.
        LINKS
            gives a list of links whose results are to be reported. The default is NONE.
        LID
            specifies that the LID control Name in subcatchment Subcatch should have a
            detailed performance report for it written to file Fname.

        The SUBCATCHMENTS, NODES, LINKS, and LID lines can be repeated multiple times.
    """
    _label = REPORT

    class KEYS:
        INPUT = 'INPUT'
        CONTINUITY = 'CONTINUITY'
        FLOWSTATS = 'FLOWSTATS'
        CONTROLS = 'CONTROLS'
        SUBCATCHMENTS = 'SUBCATCHMENTS'
        NODES = 'NODES'
        LINKS = 'LINKS'
        LID = 'LID'

    @classmethod
    def from_inp_lines(cls, lines):
        data = cls()
        for key, *line in line_iter(lines):
            key = key.upper()
            if len(line) == 1:
                value = infer_type(line[0])

            elif (key == cls.KEYS.LID) and (len(line) == 3):
                value = {'Name': line[0],
                         'Subcatch': line[1],
                         'Fname': line[2]}

            else:
                value = infer_type(line)

            if key in [cls.KEYS.SUBCATCHMENTS,
                       cls.KEYS.NODES,
                       cls.KEYS.LINKS,
                       cls.KEYS.LID]:
                if isinstance(value, str) and (value.upper() == 'ALL'):
                    pass
                elif value is None:
                    pass
                elif not isinstance(value, list):
                    value = [value]
            if key not in data:
                data[key] = value
            elif isinstance(data[key], list):
                data[key] += value
            else:
                data[key] = value
        return data

    def to_inp_lines(self, fast=False, sort_objects_alphabetical=False):
        f = ''
        max_len = len(max(self.keys(), key=len)) + 2

        def _dict_format(key, value):
            return '{key}{value}'.format(key=key.ljust(max_len),
                                         value=type2str(value) + '\n')

        for sub in self:
            value = self[sub]
            if value is None:
                continue

            if isinstance(value, list) and len(value) > 20:
                size = len(value)
                start = 0
                for end in range(20, size, 20):
                    f += _dict_format(key=sub, value=value[start:end])
                    start = end

            else:
                f += _dict_format(key=sub, value=value)

        return f


class EvaporationSection(InpSectionGeneric):
    """
    Section: [**EVAPORATION**]

    Purpose:
        Specifies how daily evaporation rates vary with time for the study area.

    Formats:
        ::

            CONSTANT    evap (0)
            MONTHLY     e1 e2 e3 e4 e5 e6 e7 e8 e9 e10 e11 e12
            TIMESERIES  Tseries
            TEMPERATURE
            FILE        (p1 p2 p3 p4 p5 p6 p7 p8 p9 p10 p11 p12)

            RECOVERY    patternID
            DRY_ONLY    NO / YES

    Remarks:
        evap
             constant evaporation rate (in/day or mm/day).
        e1
             evaporation rate in January (in/day or mm/day).
        `...`
            `...`
        e12
             evaporation rate in December (in/day or mm/day).
        Tseries
             name of time series in [TIMESERIES] section with evaporation data.
        p1
             pan coefficient for January.
        `...`
            `...`
        p12
             pan coefficient for December.
        patID
             name of a monthly time pattern.

        Use only one of the above formats (CONSTANT, MONTHLY, TIMESERIES,
        TEMPERATURE, or FILE). If no [EVAPORATION] section appears, then evaporation is
        assumed to be 0.

        TEMPERATURE indicates that evaporation rates will be computed from the daily air
        temperatures contained in an external climate file whose name is provided in the
        [TEMPERATURE] section (see below). This method also uses the site’s latitude, which
        can also be specified in the [TEMPERATURE] section.

        FILE indicates that evaporation data will be read directly from the same external
        climate file used for air temperatures as specified in the [TEMPERATURE] section
        (see below).

        RECOVERY identifies an optional monthly time pattern of multipliers used to modify
        infiltration recovery rates during dry periods. For example, if the normal infiltration
        recovery rate was 1% during a specific time period and a pattern factor of 0.8 applied
        to this period, then the actual recovery rate would be 0.8%.

        DRY_ONLY determines if evaporation only occurs during periods with no precipitation.
        The default is NO.
    """
    _label = EVAPORATION

    class KEYS:
        CONSTANT = 'CONSTANT'
        MONTHLY = 'MONTHLY'
        TIMESERIES = 'TIMESERIES'
        TEMPERATURE = 'TEMPERATURE'
        FILE = 'FILE'
        RECOVERY = 'RECOVERY'
        DRY_ONLY = 'DRY_ONLY'

        _possible = [CONSTANT, MONTHLY, TIMESERIES, TEMPERATURE, FILE, RECOVERY, DRY_ONLY]

    @classmethod
    def from_inp_lines(cls, lines):
        """
        read ``.inp``-file lines and create an section object

        Args:
            lines (str | list[list[str]]): lines in the section of the ``.inp``-file

        Returns:
            InpSectionGeneric: object of the Evaporation-section
        """
        data = cls()
        for key, *line in line_iter(lines):
            key = key.upper()
            if len(line) == 1:
                value = line[0]

            elif key == cls.KEYS.TEMPERATURE:
                assert len(line) == 0
                value = ''

            elif key == cls.KEYS.MONTHLY:
                assert len(line) == 12
                value = line

            elif key == cls.KEYS.FILE:
                if len(line) == 12:
                    value = line
                elif len(line) == 0:
                    value = ''
                else:
                    raise NotImplementedError()

            else:
                value = line

            data[key] = infer_type(value)

        mult_infos = [x in data for x in [cls.KEYS.CONSTANT, cls.KEYS.MONTHLY, cls.KEYS.TIMESERIES,
                                          cls.KEYS.TEMPERATURE, cls.KEYS.FILE]]

        if sum(mult_infos) != 1:
            if sum(mult_infos) == 0:
                data[cls.KEYS.CONSTANT] = 0
            else:
                raise UserWarning('Too much evaporation')

        return data


class TemperatureSection(InpSectionGeneric):
    """
    Section: [**TEMPERATURE**]

    Purpose:
        Specifies daily air temperatures, monthly wind speed, and various snowmelt
        parameters for the study area. Required only when snowmelt is being modeled or
        when evaporation rates are computed from daily temperatures or are read from an
        external climate file.

    Formats:
        ::

            TIMESERIES Tseries
            FILE Fname (Start)
            WINDSPEED MONTHLY s1 s2 s3 s4 s5 s6 s7 s8 s9 s10 s11 s12
            WINDSPEED FILE
            SNOWMELT Stemp ATIwt RNM Elev Lat DTLong
            ADC IMPERVIOUS f.0 f.1 f.2 f.3 f.4 f.5 f.6 f.7 f.8 f.9
            ADC PERVIOUS f.0 f.1 f.2 f.3 f.4 f.5 f.6 f.7 f.8 f.9

    Remarks:
        Tseries
            name of time series in ``[TIMESERIES]`` section with temperature data.
        Fname
            name of external Climate file with temperature data.
        Start
            date to begin reading from the file in month/day/year format (default is the beginning of the file).
        s1
            average wind speed in January (mph or km/hr).
        `...`
            `...`
        s12
            average wind speed in December (mph or km/hr).
        Stemp
            air temperature at which precipitation falls as snow (deg F or C).
        ATIwt
            antecedent temperature index weight (default is 0.5).
        RNM
            negative melt ratio (default is 0.6).
        Elev
            average elevation of study area above mean sea level (ft or m) (default is 0).
        Lat
            latitude of the study area in degrees North (default is 50).
        DTLong
            correction, in minutes of time, between true solar time and the standard clock time (default is 0).
        f.0
            fraction of area covered by snow when ratio of snow depth to depth at 100% cover is 0
        `...`
            `...`
        f.9
            fraction of area covered by snow when ratio of snow depth to depth at 100% cover is 0.9

    Use the ``TIMESERIES`` line to read air temperature from a time series or the ``FILE`` line
    to read it from an external Climate file.
    Climate files are discussed in Section 11.4 Climate Files.
    If neither format is used, then air temperature remains constant at 70 degrees F (=21.11111 °C).

    Wind speed can be specified either by monthly average values or by the same
    Climate file used for air temperature. If neither option appears, then wind speed is
    assumed to be 0.

    Separate Areal Depletion Curves (ADC) can be defined for impervious and pervious
    sub-areas. The ADC parameters will default to 1.0 (meaning no depletion) if no data
    are supplied for a particular type of sub-area.
    """
    _label = TEMPERATURE

    class KEYS:
        TIMESERIES = 'TIMESERIES'
        FILE = 'FILE'

        _WINDSPEED = 'WINDSPEED'
        WINDSPEED_MONTHLY = _WINDSPEED + ' MONTHLY'
        WINDSPEED_FILE = _WINDSPEED + ' FILE'

        SNOWMELT = 'SNOWMELT'

        _ADC = 'ADC'
        ADC_IMPERVIOUS = _ADC + ' IMPERVIOUS'
        ADC_PERVIOUS = _ADC + ' PERVIOUS'

    @classmethod
    def from_inp_lines(cls, lines):
        """
        read ``.inp``-file lines and create an section object

        Args:
            lines (str | list[list[str]]): lines in the section of the ``.inp``-file

        Returns:
            InpSectionGeneric: object of the Temperature-section
        """
        data = cls()
        for key, *line in line_iter(lines):
            key = key.upper()
            n_options = len(line)

            if key == cls.KEYS.TIMESERIES:
                assert n_options == 1
                value = line[0]

            elif key == cls.KEYS.FILE:
                if n_options == 1:
                    value = line[0]
                else:
                    value = line

            elif key == cls.KEYS._WINDSPEED:
                key += ' ' + line.pop(0)
                if key == cls.KEYS.WINDSPEED_FILE:
                    assert n_options == 1
                    value = ''
                elif key == cls.KEYS.WINDSPEED_MONTHLY:
                    assert n_options == 13
                    value = line
                else:
                    raise NotImplementedError()

            elif key == cls.KEYS.SNOWMELT:
                assert n_options == 6
                value = line

            elif key == cls.KEYS._ADC:
                key += ' ' + line.pop(0)
                assert n_options == 11
                value = line
                if key not in [cls.KEYS.ADC_IMPERVIOUS, cls.KEYS.ADC_PERVIOUS]:
                    raise NotImplementedError()

            else:
                raise NotImplementedError()

            data[key] = value

        return data


class MapSection(InpSectionGeneric):
    """
    Section: [**MAP**]

    Purpose:
        Provides dimensions and distance units for the map.

    Formats:
        ::

            DIMENSIONS X1 Y1 X2 Y2
            UNITS FEET / METERS / DEGREES / NONE

    Args:
        dimensions (list[float, float, float, float]): lower-left and upper-right coordinates of the full map extent

            - lower_left_x (:obj:`float`): lower-left X coordinate ``X1``
            - lower_left_y (:obj:`float`): lower-left Y coordinate ``Y1``
            - upper_right_x (:obj:`float`): upper-right X coordinate ``X2``
            - upper_right_y (:obj:`float`): upper-right Y coordinate ``Y2``
        units (str): one of FEET / METERS / DEGREES / NONE see :py:attr:`~MapSection.UNITS`

    Attributes:
        lower_left_x (float): lower-left X coordinate ``X1``
        lower_left_y (float): lower-left Y coordinate ``Y1``
        upper_right_x (float): upper-right X coordinate ``X2``
        upper_right_y (float): upper-right Y coordinate ``Y2``
        units (str): one of FEET / METERS / DEGREES / NONE see :py:attr:`~MapSection.UNITS`
    """
    _label = MAP

    class KEYS:
        DIMENSIONS = 'DIMENSIONS'
        UNITS = 'UNITS'

    class UNIT_OPTIONS:
        FEET = 'FEET'
        METERS = 'METERS'
        DEGREES = 'DEGREES'
        NONE = None

    @classmethod
    def from_inp_lines(cls, lines):
        """
        read ``.inp``-file lines and create an section object

        Args:
            lines (str | list[list[str]]): lines in the section of the ``.inp``-file

        Returns:
            InpSectionGeneric: object of the Map-section
        """
        data = cls()
        for key, *line in line_iter(lines):
            key = key.upper()
            if key == cls.KEYS.DIMENSIONS:
                data[key] = [float(i) for i in line]

            elif key == cls.KEYS.UNITS:
                data[key] = line[0]
            else:
                raise NotImplementedError()
        return data


class FilesSection(InpSectionGeneric):
    """
    Section: [**FILES**]

    Purpose:
        Identifies optional interface files used or saved by a run.

    Format:
        ::

            USE / SAVE RAINFALL Fname
            USE / SAVE RUNOFF Fname
            USE / SAVE HOTSTART Fname
            USE / SAVE RDII Fname
            USE INFLOWS Fname
            SAVE OUTFLOWS Fname

    Remarks:
        Fname
            is the name of an interface file.

        Refer to Section 11.7 Interface Files for a description of interface files. Rainfall, Runoff, and
        RDII files can either be used or saved in a run, but not both. A run can both use and save a Hot
        Start file (with different names).
    """
    _label = FILES

    class KEYS:
        USE = 'USE'
        SAVE = 'SAVE'

        _use_or_save = [USE, SAVE]

        RAINFALL = 'RAINFALL'
        RUNOFF = 'RUNOFF'
        HOTSTART = 'HOTSTART'
        RDII = 'RDII'
        INFLOWS = 'INFLOWS'
        OUTFLOWS = 'OUTFLOWS'

        _possible = [RAINFALL, RUNOFF, HOTSTART, RDII, INFLOWS, OUTFLOWS]

    def __setitem__(self, key, item):
        InpSectionGeneric.__setitem__(self, key, item)

    def __delitem__(self, key):
        InpSectionGeneric.__delitem__(self, key)

    @classmethod
    def from_inp_lines(cls, lines):
        """
        read ``.inp``-file lines and create an section object

        Args:
            lines (str | list[list[str]]): lines in the section of the ``.inp``-file

        Returns:
            InpSectionGeneric: object of the Files-section
        """
        data = cls()
        for use_save, kind, *fn in line_iter(lines):
            use_save = use_save.upper()
            kind = kind.upper()
            assert use_save in cls.KEYS._use_or_save
            assert kind in cls.KEYS._possible
            data[f'{use_save} {kind}'] = ' '.join(fn).strip('"')
        return data

    def use(self, kind, filename):
        self[f'{self.KEYS.USE} {kind}'] = filename

    def save(self, kind, filename):
        self[f'{self.KEYS.SAVE} {kind}'] = filename


class AdjustmentsSection(InpSectionGeneric):
    """
    Section: [**ADJUSTMENTS**]

    Purpose:
        Specifies optional monthly adjustments to be made to temperature, evaporation rate,
        rainfall intensity and hydraulic conductivity in each time period of a simulation.

    Format:
        ::

            TEMPERATURE  t1 t2 t3 t4 t5 t6 t7 t8 t9 t10 t11 t12
            EVAPORATION  e1 e2 e3 e4 e5 e6 e7 e8 e9 e10 e11 e12
            RAINFALL     r1 r2 r3 r4 r5 r6 r7 r8 r9 r10 r11 r12
            CONDUCTIVITY c1 c2 c3 c4 c5 c6 c7 c8 c9 c10 c11 c12

    Remarks:
        t1..t12
            adjustments to temperature in January, February, etc., as plus or minus degrees F (degrees C).
        e1..e12
            adjustments to evaporation rate in January, February, etc., as plus or minus in/day (mm/day).
        r1..r12
            multipliers applied to precipitation rate in January, February, etc.
        c1..c12
            multipliers applied to soil hydraulic conductivity in January, February, etc. used in either Horton or
            Green-Ampt infiltration.

        The same adjustment is applied for each time period within a given month and is repeated for that
        month in each subsequent year being simulated.
    """
    _label = ADJUSTMENTS

    class KEYS:
        TEMPERATURE = 'TEMPERATURE'
        EVAPORATION = 'EVAPORATION'
        RAINFALL = 'RAINFALL'
        CONDUCTIVITY = 'CONDUCTIVITY'

        _possible = [TEMPERATURE, EVAPORATION, RAINFALL, CONDUCTIVITY]

    @classmethod
    def from_inp_lines(cls, lines):
        """
        read ``.inp``-file lines and create an section object

        Args:
            lines (str | list[list[str]]): lines in the section of the ``.inp``-file

        Returns:
            InpSectionGeneric: object of the Adjustments-section
        """
        data = cls()
        for key, *factors in line_iter(lines):
            key = key.upper()
            assert len(factors) == 12
            assert key in cls.KEYS._possible
            data[key] = [float(i) for i in factors]
        return data


class BackdropSection(InpSectionGeneric):
    """
    Section: [**BACKDROP**]

    Purpose:
        Specifies file name and coordinates of map’s backdrop image.

    Format:
        ::

            TFILE        Fname
            DIMENSIONS   X1 Y1 X2 Y2

    Remarks:
        Fname
            name of file containing backdrop image
        X1
            lower-left X coordinate of backdrop image
        Y1
            lower-left Y coordinate of backdrop image
        X2
            upper-right X coordinate of backdrop image
        Y2
            upper-right Y coordinate of backdrop image
    """
    _label = BACKDROP

    class KEYS:
        FILE = 'FILE'
        DIMENSIONS = 'DIMENSIONS'
        UNITS = 'UNITS'  # not in documentation
        OFFSET = 'OFFSET'  # not in documentation
        SCALING = 'SCALING'  # not in documentation

    @classmethod
    def from_inp_lines(cls, lines):
        """
        read ``.inp``-file lines and create an section object

        Args:
            lines (str | list[list[str]]): lines in the section of the ``.inp``-file

        Returns:
            InpSectionGeneric: object of the Backdrop-section
        """
        data = cls()
        for key, *line in line_iter(lines):
            key = key.upper()
            if key == cls.KEYS.FILE:
                data[key] = ' '.join(line)
            elif key == cls.KEYS.DIMENSIONS:
                assert len(line) == 4
                data[key] = [float(i) for i in line]
            elif key in [cls.KEYS.UNITS, cls.KEYS.OFFSET, cls.KEYS.SCALING]:
                # unknown behavior not in documentation
                data[key] = ' '.join(line)
            else:
                raise NotImplementedError()
        return data



from datetime import timedelta
from math import floor

from swmm.toolkit import solver
from pyswmm import Simulation
from tqdm.auto import tqdm
from swmm_api.run import get_result_filenames


def run(fn_inp):
    solver.swmm_run(fn_inp, *get_result_filenames(fn_inp))


def run_progress(fn_inp, n_total=100):
    with Simulation(fn_inp) as sim:
        total_time_seconds = (sim.end_time - sim.start_time) / timedelta(seconds=1)
        sim.step_advance(floor(total_time_seconds / n_total))
        progress = tqdm(total=n_total, desc=f'swmm5 {fn_inp}')
        for _ in sim:
            progress.update(1)
            progress.postfix = f'{sim.current_time}'
        progress.update(progress.total - progress.n)
        progress.postfix = f'{sim.current_time}'
        progress.close()


def get_swmm_version():
    return '.'.join(solver.swmm_version_info())



from numpy import NaN

from ._identifiers import IDENTIFIERS
from ..helpers import BaseSectionObject
from ..section_labels import LID_USAGE, LID_CONTROLS


class LIDControl(BaseSectionObject):
    """
    Section: [**LID_CONTROLS**]

    Purpose:
        Defines scale-independent LID controls that can be deployed within subcatchments.

    Formats:
        ::

            Name SURFACE StorHt VegFrac Rough Slope Xslope
            Name SOIL Thick Por FC WP Ksat Kcoeff Suct
            Name PAVEMENT Thick Vratio FracImp Perm Vclog
            Name STORAGE Height Vratio Seepage Vclog
            Name DRAIN Coeff Expon Offset Delay
            Name DRAINMAT Thick Vratio Rough

    Attributes:
        Name (str):
            name assigned to LID process.
        lid_kind (str): - BC for bio-retention cell
                        - RG for rain garden; GR for green roof
                        - IT for infiltration trench
                        - PP for permeable pavement
                        - RB for rain barrel
                        - RD for rooftop disconnection
                        - VS for vegetative swale.

    Examples:
        ::

            ;A street planter with no drain
            Planter BC
            Planter SURFACE 6 0.3 0 0 0
            Planter SOIL 24 0.5 0.1 0.05 1.2 2.4
            Planter STORAGE 12 0.5 0.5 0

            ;A green roof with impermeable bottom
            GR1 BC
            GR1 SURFACE 3 0 0 0 0
            GR1 SOIL 3 0.5 0.1 0.05 1.2 2.4
            GR1 STORAGE 3 0.5 0 0
            GR1 DRAIN 5 0.5 0 0

            ;A rain barrel that drains 6 hours after rainfall ends
            RB12 RB
            RB12 STORAGE 36 0 0 0
            RB12 DRAIN 10 0.5 0 6

            ;A grass swale 24 in. high with 5:1 side slope
            Swale VS
            Swale SURFACE 24 0 0.2 3 5

    Remarks:
        The following table shows which layers are required (x) or are optional (o) for each type of LID process:

        +-----------------------+---------+----------+------+---------+-------+-----------+
        | LID Type              | Surface | Pavement | Soil | Storage | Drain | Drain Mat |
        +-----------------------+---------+----------+------+---------+-------+-----------+
        | Bio-Retention Cell    | x       |          | x    | x       | o     |           |
        +-----------------------+---------+----------+------+---------+-------+-----------+
        | Rain Garden           | x       |          | x    |         |       |           |
        +-----------------------+---------+----------+------+---------+-------+-----------+
        | Green Roof            | x       |          | x    |         |       | x         |
        +-----------------------+---------+----------+------+---------+-------+-----------+
        | Infiltration Trench   | x       |          |      | x       | o     |           |
        +-----------------------+---------+----------+------+---------+-------+-----------+
        | Permeable Pavement    | x       | x        | o    | x       | o     |           |
        +-----------------------+---------+----------+------+---------+-------+-----------+
        | Rain Barrel           |         |          |      | x       | x     |           |
        +-----------------------+---------+----------+------+---------+-------+-----------+
        | Rooftop Disconnection | x       |          |      |         | x     |           |
        +-----------------------+---------+----------+------+---------+-------+-----------+
        | Vegetative Swale      | x       |          |      |         |       |           |
        +-----------------------+---------+----------+------+---------+-------+-----------+


        The equation used to compute flow rate out of the underdrain per unit area of the LID (in in/hr or
        mm/hr) is q = C * (h - H_d) ^ n where q is outflow, h is height of stored water (inches or mm) and H d
        is the drain offset height. Note that the units of C depend on the unit system being used as well
        as the value assigned to n.

        The actual dimensions of an LID control are provided in the [LID_USAGE] section when it is
        placed in a particular subcatchment.
    """
    _identifier = (IDENTIFIERS.Name, 'lid_kind')
    _section_label = LID_CONTROLS
    _table_inp_export = False

    class LID_TYPES:
        BC = 'BC'  # for bio-retention cell
        RG = 'RG'  # for rain garden
        GR = 'GR'  # for green roof
        IT = 'IT'  # for infiltration trench
        PP = 'PP'  # for permeable pavement
        RB = 'RB'  # for rain barrel
        RD = 'RD'  # for rooftop disconnection
        VS = 'VS'  # for vegetative swale.

        _possible = [BC, RG, GR, IT, PP, RB, RD, VS]

    def __init__(self, Name, lid_kind, layer_dict=None):
        """
        Create LID_CONTROLS object

        Args:
            Name (str):
            name assigned to LID process.
        lid_kind (str): - BC for bio-retention cell
                        - RG for rain garden; GR for green roof
                        - IT for infiltration trench
                        - PP for permeable pavement
                        - RB for rain barrel
                        - RD for rooftop disconnection
                        - VS for vegetative swale.
            layer_dict (dict[str, LIDControl.LAYER_TYPES.Surface]): dict of used layers in control
        """
        self.Name = str(Name)
        self.lid_kind = lid_kind.upper()  # one of LID_TYPES
        if layer_dict is None:
            self.layer_dict = dict()
        else:
            self.layer_dict = layer_dict

    @classmethod
    def _convert_lines(cls, multi_line_args):
        last = None

        for name, *line in multi_line_args:
            # ---------------------------------
            if line[0].upper() in cls.LID_TYPES._possible:
                if last is not None:
                    yield last
                last = cls(name, lid_kind=line[0].upper())
            elif name == last.Name:
                surface_kind = line.pop(0).upper()  # one of SURFACE_TYPES
                last.layer_dict[surface_kind] = cls.LAYER_TYPES._dict[surface_kind](*line)
        yield last

    class LAYER_TYPES:
        class Surface(BaseSectionObject):
            _LABEL = 'SURFACE'

            def __init__(self, StorHt, VegFrac, Rough, Slope, Xslope):
                """
                Args:
                    StorHt (float): when confining walls or berms are present this is the maximum depth to which water can
                        pond above the surface of the unit before overflow occurs (in inches or mm). For LIDs that
                        experience overland flow it is the height of any surface depression storage.
                        For swales, it is the height of its trapezoidal cross section.
                    VegFrac (float): fraction of the surface storage volume that is filled with vegetation.
                    Rough (float): Manning's n for overland flow over surface soil cover, pavement, roof surface or a
                        vegetative swale. Use 0 for other types of LIDs.
                    Slope (float): slope of a roof surface, pavement surface or vegetative swale (percent).
                        Use 0 for other types of LIDs.
                    Xslope (float): slope (run over rise) of the side walls of a vegetative swale's cross section.
                        Use 0 for other types of LIDs.

                Remarks:
                    If either Rough or Slope values are 0 then any ponded water that exceeds the
                    surface storage depth is assumed to completely overflow the LID control within a
                    single time step.
                """
                self.StorHt = float(StorHt)
                self.VegFrac = float(VegFrac)
                self.Rough = float(Rough)
                self.Slope = float(Slope)
                self.Xslope = float(Xslope)

        class Soil(BaseSectionObject):
            _LABEL = 'SOIL'

            def __init__(self, Thick, Por, FC, WP, Ksat, Kcoeff, Suct):
                """
                Args:
                    Thick (float): thickness of the soil layer (inches or mm).
                    Por (float): soil porosity (volume of pore space relative to total volume).
                    FC (float): soil field capacity (volume of pore water relative to total volume after the
                        soil has been allowed to drain fully).
                    WP (float): soil wilting point (volume of pore water relative to total volume for a well
                        dried soil where only bound water remains).
                    Ksat (float): soil’s saturated hydraulic conductivity (in/hr or mm/hr).
                    Kcoeff (float): slope of the curve of log(conductivity) versus soil moisture content
                        (dimensionless).
                    Suct (float): soil capillary suction (in or mm).
                """
                self.Thick = float(Thick)
                self.Por = float(Por)
                self.FC = float(FC)
                self.WP = float(WP)
                self.Ksat = float(Ksat)
                self.Kcoeff = float(Kcoeff)
                self.Suct = float(Suct)

        class Pavement(BaseSectionObject):
            _LABEL = 'PAVEMENT'

            def __init__(self, Thick, Vratio, FracImp, Perm, Vclog, regeneration_interval=NaN, regeneration_fraction=NaN):
                """
                Args:
                    Thick (float): thickness of the pavement layer (inches or mm).
                    Vratio (float): void ratio (volume of void space relative to the volume of solids in the
                        pavement for continuous systems or for the fill material used in modular
                        systems). Note that porosity = void ratio / (1 + void ratio).

                    FracImp (float): ratio of impervious paver material to total area for modular systems; 0 for
                        continuous porous pavement systems.
                    Perm (float): permeability of the concrete or asphalt used in continuous systems or
                        hydraulic conductivity of the fill material (gravel or sand) used in modular
                        systems (in/hr or mm/hr).
                    Vclog (float): number of pavement layer void volumes of runoff treated it takes to
                        completely clog the pavement. Use a value of 0 to ignore clogging.
                """
                self.Thick = float(Thick)
                self.Vratio = float(Vratio)
                self.FracImp = float(FracImp)
                self.Perm = float(Perm)
                self.Vclog = Vclog  # acc. to documentation
                # self.clogging_factor = float(clogging_factor)
                self.regeneration_interval = float(regeneration_interval)
                self.regeneration_fraction = float(regeneration_fraction)

        class Storage(BaseSectionObject):
            _LABEL = 'STORAGE'

            def __init__(self, Height, Vratio, Seepage, Vclog):
                """
                Args:
                    Height (float): thickness of the storage layer or height of a rain barrel (inches or mm).
                    Vratio (float): void ratio (volume of void space relative to the volume of solids in the
                        layer). Note that porosity = void ratio / (1 + void ratio).
                    Seepage (float): the rate at which water seeps from the layer into the underlying native
                        soil when first constructed (in/hr or mm/hr). If there is an impermeable
                        floor or liner below the layer then use a value of 0.
                    Vclog (int): number of storage layer void volumes of runoff treated it takes to
                        completely clog the layer. Use a value of 0 to ignore clogging.

                Remarks:
                    Values for Vratio, Seepage, and Vclog are ignored for rain barrels.
                """
                self.Height = float(Height)
                self.Vratio = float(Vratio)
                self.Seepage = float(Seepage)
                self.Vclog = int(Vclog)

        class Drain(BaseSectionObject):
            _LABEL = 'DRAIN'

            def __init__(self, Coeff, Expon, Offset, Delay, open_level=NaN, close_level=NaN):
                """
                Args:
                    Coeff (float): coefficient C that determines the rate of flow through the drain as a
                        function of height of stored water above the drain bottom. For Rooftop
                        Disconnection it is the maximum flow rate (in inches/hour or mm/hour)
                        that the roof’s gutters and downspouts can handle before overflowing.
                    Expon (float): exponent n that determines the rate of flow through the drain as a
                        function of height of stored water above the drain outlet.
                    Offset (float): height of the drain line above the bottom of the storage layer or rain
                        barrel (inches or mm).
                    Delay: number of dry weather hours that must elapse before the drain line in a
                        rain barrel is opened (the line is assumed to be closed once rainfall
                        begins). A value of 0 signifies that the barrel's drain line is always open
                        and drains continuously. This parameter is ignored for other types of
                        LIDs.
                """
                self.Coeff = float(Coeff)
                self.Expon = float(Expon)
                self.Offset = float(Offset)
                self.Delay = Delay  # acc. to documentation  / for Rain Barrels only
                # self.open_level = open_level  # to in documentation
                # self.close_level = close_level  # to in documentation

        class Drainmat(BaseSectionObject):
            _LABEL = 'DRAINMAT'

            def __init__(self, Thick, Vratio, Rough):
                """
                Args:
                    Thick (float): thickness of the drainage mat (inches or mm).
                    Vratio (float): ratio of void volume to total volume in the mat.
                    Rough (float): Manning's n constant used to compute the horizontal flow rate of drained water through the mat.
                """
                self.Thick = float(Thick)
                self.Vratio = float(Vratio)
                self.Rough = float(Rough)

        SURFACE = Surface._LABEL
        SOIL = Soil._LABEL
        PAVEMENT = Pavement._LABEL
        STORAGE = Storage._LABEL
        DRAIN = Drain._LABEL
        DRAINMAT = Drainmat._LABEL

        _possible = [SURFACE, SOIL, PAVEMENT, STORAGE, DRAIN, DRAINMAT]

        _dict = {x._LABEL: x for x in (Surface, Soil, Pavement, Storage, Drain, Drainmat)}

    def to_inp_line(self):
        s = '{} {}\n'.format(self.Name, self.lid_kind)
        for layer, l in self.layer_dict.items():
            s += '{} {:<8} '.format(self.Name, layer) + l.to_inp_line() + '\n'
        return s


class LIDUsage(BaseSectionObject):
    """
    Section: [**LID_USAGE**]

    Purpose:
        Deploys LID controls within specific subcatchment areas.

    Formats:
        ::

            Subcat LID Number Area Width InitSat FromImp ToPerv (RptFile DrainTo)

    Args:
        Subcat (str): name of the subcatchment using the LID process.
        LID (str): name of an LID process defined in the [LID_CONTROLS] section.
        Number (int): number of replicate LID units deployed.
        Area (float): area of each replicate unit (ft^2 or m^2 ).
        Width (float): width of the outflow face of each identical LID unit (in ft or m). This
            parameter applies to roofs, pavement, trenches, and swales that use
            overland flow to convey surface runoff off of the unit. It can be set to 0 for
            other LID processes, such as bio-retention cells, rain gardens, and rain
            barrels that simply spill any excess captured runoff over their berms.
        InitSat (float): for bio-retention cells, rain gardens, and green roofs this is the degree to
            which the unit's soil is initially filled with water (0 % saturation
            corresponds to the wilting point moisture content, 100 % saturation has
            the moisture content equal to the porosity). The storage zone beneath
            the soil zone of the cell is assumed to be completely dry. For other types
            of LIDs it corresponds to the degree to which their storage zone is
            initially filled with water
        FromImp (float): percent of the impervious portion of the subcatchment’s non-LID area
            whose runoff is treated by the LID practice. (E.g., if rain barrels are used
            to capture roof runoff and roofs represent 60% of the impervious area,
            then the impervious area treated is 60%). If the LID unit treats only direct
            rainfall, such as with a green roof, then this value should be 0. If the LID
            takes up the entire subcatchment then this field is ignored.
        ToPerv (int): a value of 1 indicates that the surface and drain flow from the LID unit
            should be routed back onto the pervious area of the subcatchment that
            contains it. This would be a common choice to make for rain barrels,
            rooftop disconnection, and possibly green roofs. The default value is 0.
        RptFile (str): optional name of a file to which detailed time series results for the LID
            will be written. Enclose the name in double quotes if it contains spaces
            and include the full path if it is different than the SWMM input file path.
            Use ‘*’ if not applicable and an entry for DrainTo follows
        DrainTo (str): optional name of subcatchment or node that receives flow from the unit’s
            drain line, if different from the outlet of the subcatchment that the LID is
            placed in.

    Remarks:
        If ``ToPerv`` is set to 1 and ``DrainTo`` set to some other outlet, then only the excess
        surface flow from the LID unit will be routed back to the subcatchment’s pervious
        area while the underdrain flow will be sent to ``DrainTo``.

        More than one type of LID process can be deployed within a subcatchment as long
        as their total area does not exceed that of the subcatchment and the total percent
        impervious area treated does not exceed 100.

    Examples:
        ::

            ;34 rain barrels of 12 sq ft each are placed in
            ;subcatchment S1. They are initially empty and treat 17%
            ;of the runoff from the subcatchment’s impervious area.
            ;The outflow from the barrels is returned to the
            ;subcatchment’s pervious area.
            S1 RB14 34 12 0 0 17 1

            ;Subcatchment S2 consists entirely of a single vegetative
            ;swale 200 ft long by 50 ft wide.
            S2 Swale 1 10000 50 0 0 0 “swale.rpt”
    """
    _identifier = (IDENTIFIERS.Subcatch, 'LID')
    _section_label = LID_USAGE

    def __init__(self, Subcatch, LID, Number, Area, Width, InitSat, FromImp, ToPerv, RptFile=NaN, DrainTo=NaN):
        self.Subcatch = str(Subcatch)
        self.LID = str(LID)
        self.Number = Number
        self.Area = float(Area)
        self.Width = float(Width)
        self.InitSat = float(InitSat)
        self.FromImp = float(FromImp)
        self.ToPerv = int(ToPerv)
        self.RptFile = RptFile
        self.DrainTo = DrainTo



from numpy import NaN
from pandas import DataFrame

from .._type_converter import GIS_FLOAT_FORMAT
from ..helpers import BaseSectionObject, SWMM_VERSION
from ._identifiers import IDENTIFIERS
from ..section_labels import SUBCATCHMENTS, SUBAREAS, INFILTRATION, POLYGONS, LOADINGS, COVERAGES, GWF, GROUNDWATER


class SubCatchment(BaseSectionObject):
    """
    Section: [**SUBCATCHMENTS**]

    Purpose:
        Identifies each sub-catchment within the study area. Subcatchments are land area
        units which generate runoff from rainfall.

    Format:
        ::

            Name Rgage OutID Area %Imperv Width Slope Clength (Spack)

    Format-PCSWMM:
        ``Name RainGage Outlet Area %Imperv Width %Slope CurbLen SnowPack``

    Args:
        Name (str): name assigned to subcatchment.
        RainGage (str): name of rain gage in [RAINGAGES] section assigned to subcatchment. ``Rgage``
        Outlet (str): name of node or subcatchment that receives runoff from subcatchment. ``OutID``
        Area (float): area of subcatchment (acres or hectares).
        Imperv (float): percent imperviousness of subcatchment. ``%Imperv``
        Width (float): characteristic width of subcatchment (ft or meters).
        Slope (float): subcatchment slope (percent).
        CurbLen (float): total curb length (any length units). Use 0 if not applicable. ``Clength``
        SnowPack (str): optional name of snow pack object (from [SNOWPACKS] section) that characterizes snow
        accumulation and melting over the subcatchment. ``Spack``

    Attributes:
        Name (str): name assigned to subcatchment.
        RainGage (str): name of rain gage in [RAINGAGES] section assigned to subcatchment. ``Rgage``
        Outlet (str): name of node or subcatchment that receives runoff from subcatchment. ``OutID``
        Area (float): area of subcatchment (acres or hectares).
        Imperv (float): percent imperviousness of subcatchment. ``%Imperv``
        Width (float): characteristic width of subcatchment (ft or meters).
        Slope (float): subcatchment slope (percent).
        CurbLen (float): total curb length (any length units). Use 0 if not applicable. ``Clength``
        SnowPack (str): optional name of snow pack object (from [SNOWPACKS] section) that characterizes snow
        accumulation and melting over the subcatchment. ``Spack``
    """
    _identifier = IDENTIFIERS.Name
    _section_label = SUBCATCHMENTS

    def __init__(self, Name, RainGage, Outlet, Area, Imperv, Width, Slope, CurbLen=0, SnowPack=NaN):
        self.Name = str(Name)
        self.RainGage = str(RainGage)
        self.Outlet = str(Outlet)
        self.Area = float(Area)
        self.Imperv = float(Imperv)
        self.Width = float(Width)
        self.Slope = float(Slope)
        self.CurbLen = float(CurbLen)
        self.SnowPack = SnowPack


class SubArea(BaseSectionObject):
    """
    Section: [**SUBAREAS**]

    Purpose:
        Supplies information about pervious and impervious areas for each subcatchment.
        Each subcatchment can consist of a pervious sub-area, an impervious sub-area with
        depression storage, and an impervious sub-area without depression storage.

    Format:
        ::

            Subcat Nimp Nperv Simp Sperv %Zero RouteTo (%Routed)

    Format-PCSWMM:
        ``Subcatchment N-Imperv N-Perv S-Imperv S-Perv PctZero RouteTo PctRouted``

    Args:
        Subcatch (str): subcatchment name. ``Subcat``
        N_Imperv (float): Manning's n for overland flow over the impervious sub-area. ``Nimp``
        N_Perv (float): Manning's n for overland flow over the pervious sub-area. ``Nperv``
        S_Imperv (float): depression storage for impervious sub-area (inches or mm). ``Simp``
        S_Perv (float): depression storage for pervious sub-area (inches or mm). ``Sperv``
        PctZero (float): percent of impervious area with no depression storage. ``%Zero``
        RouteTo (str):

            - ``IMPERVIOUS`` if pervious area runoff runs onto impervious area,
            - ``PERVIOUS`` if impervious runoff runs onto pervious area,
            - ``OUTLET`` if both areas drain to the subcatchment's outlet (default = ``OUTLET``).

        PctRouted (float): percent of runoff routed from one type of area to another (default = 100). ``%Routed``

    Attributes:
        Subcatch (str): subcatchment name. ``Subcat``
        N_Imperv (float): Manning's n for overland flow over the impervious sub-area. ``Nimp``
        N_Perv (float): Manning's n for overland flow over the pervious sub-area. ``Nperv``
        S_Imperv (float): depression storage for impervious sub-area (inches or mm). ``Simp``
        S_Perv (float): depression storage for pervious sub-area (inches or mm). ``Sperv``
        PctZero (float): percent of impervious area with no depression storage. ``%Zero``
        RouteTo (str):

            - ``IMPERVIOUS`` if pervious area runoff runs onto impervious area,
            - ``PERVIOUS`` if impervious runoff runs onto pervious area,
            - ``OUTLET`` if both areas drain to the subcatchment's outlet (default = ``OUTLET``).

        PctRouted (float): percent of runoff routed from one type of area to another (default = 100). ``%Routed``
    """
    _identifier = IDENTIFIERS.Subcatch
    _section_label = SUBAREAS

    class RoutToOption:
        __class__ = 'RoutTo Option'
        IMPERVIOUS = 'IMPERVIOUS'
        PERVIOUS = 'PERVIOUS'
        OUTLET = 'OUTLET'

    def __init__(self, Subcatch, N_Imperv, N_Perv, S_Imperv, S_Perv, PctZero, RouteTo=RoutToOption.OUTLET,
                 PctRouted=100):
        self.Subcatch = str(Subcatch)
        self.N_Imperv = float(N_Imperv)
        self.N_Perv = float(N_Perv)
        self.S_Imperv = float(S_Imperv)
        self.S_Perv = float(S_Perv)
        self.PctZero = float(PctZero)
        self.RouteTo = str(RouteTo)
        self.PctRouted = float(PctRouted)


class Infiltration(BaseSectionObject):
    """
    Section: [**INFILTRATION**]

    Purpose:
        Supplies infiltration parameters for each subcatchment.
        Rainfall lost to infiltration only occurs over the pervious sub-area of a subcatchment.

    Formats:
        ::

            Subcat MaxRate MinRate Decay DryTime MaxInf
            Subcat Psi Ksat IMD
            Subcat CurveNo Ksat DryTime

    Remarks:
        Subcat
            subcatchment name.

        For Horton and Modified Horton Infiltration:
            MaxRate
                maximum infiltration rate on Horton curve (in/hr or mm/hr).
            MinRate
                minimum infiltration rate on Horton curve (in/hr or mm/hr).
            Decay
                decay rate constant of Horton curve (1/hr).
            DryTime
                time it takes for fully saturated soil to dry (days).
            MaxInf
                maximum infiltration volume possible (0 if not applicable) (in or mm).

        For Green-Ampt and Modified Green-Ampt Infiltration:
            Psi
                soil capillary suction (in or mm).
            Ksat
                soil saturated hydraulic conductivity (in/hr or mm/hr).
            IMD
             initial soil moisture deficit (volume of voids / total volume).

        For Curve-Number Infiltration:
            CurveNo
                SCS Curve Number.
            Ksat
                soil saturated hydraulic conductivity (in/hr or mm/hr)
                (This property has been deprecated and is no longer used.)
            DryTime
                time it takes for fully saturated soil to dry (days).

    Args:
        Subcatch (str): subcatchment name. ``Subcat``

    Attributes:
        Subcatch (str): subcatchment name. ``Subcat``
    """
    _identifier = IDENTIFIERS.Subcatch
    _section_label = INFILTRATION

    # _table_inp_export = False

    # jetbrains://clion/navigate/reference?project=Swmm5&path=src/infil.c : 133
    # infil_readParams

    def __init__(self, Subcatch):
        self.Subcatch = str(Subcatch)

    @classmethod
    def from_inp_line(cls, Subcatch, *args, **kwargs):

        subcls = cls

        # n_args = len(args) + len(kwargs.keys()) + 1
        # if n_args == 6:  # hortn
        #     subcls = InfiltrationHorton
        # elif n_args == 4:
        #     subcls = cls

        # _____________________________________
        sub_class_id = None
        if SWMM_VERSION == '5.1.015':
            # NEU in swmm 5.1.015
            last_arg = args[-1]
            if last_arg in INFILTRATION_DICT:
                sub_class_id = last_arg
                subcls = INFILTRATION_DICT[last_arg]
                args = args[:-1]

        if subcls != InfiltrationHorton:
            args = args[:3]

        # _____________________________________
        o = subcls(Subcatch, *args, **kwargs)
        # _____________________________________
        if sub_class_id is not None:
            o.kind = sub_class_id
        return o


class InfiltrationHorton(Infiltration):
    """
    Section: [**INFILTRATION**]

    For Horton and Modified Horton Infiltration

    Purpose:
        Supplies infiltration parameters for each subcatchment.
        Rainfall lost to infiltration only occurs over the pervious sub-area of a subcatchment.

    Formats:
        ::

            Subcat MaxRate MinRate Decay DryTime MaxInf

    Format-PCSWMM:
        ``Subcatchment MaxRate MinRate Decay DryTime MaxInfil``

    Args:
        Subcatch (str): subcatchment name. ``Subcat``
        MaxRate (float): maximum infiltration rate on Horton curve (in/hr or mm/hr).
        MinRate (float): minimum infiltration rate on Horton curve (in/hr or mm/hr).
        Decay (float): decay rate constant of Horton curve (1/hr).
        DryTime (float): time it takes for fully saturated soil to dry (days).
        MaxInf (float): maximum infiltration volume possible (0 if not applicable) (in or mm).

    Attributes:
        Subcatch (str): subcatchment name. ``Subcat``
        MaxRate (float): maximum infiltration rate on Horton curve (in/hr or mm/hr).
        MinRate (float): minimum infiltration rate on Horton curve (in/hr or mm/hr).
        Decay (float): decay rate constant of Horton curve (1/hr).
        DryTime (float): time it takes for fully saturated soil to dry (days).
        MaxInf (float): maximum infiltration volume possible (0 if not applicable) (in or mm).
    """

    def __init__(self, Subcatch, MaxRate, MinRate, Decay, DryTime, MaxInf, kind=None):
        Infiltration.__init__(self, Subcatch)
        self.MaxRate = float(MaxRate)
        self.MinRate = float(MinRate)
        self.Decay = float(Decay)
        self.DryTime = float(DryTime)
        self.MaxInf = float(MaxInf)
        self.kind = NaN


class InfiltrationGreenAmpt(Infiltration):
    """
    Section: [**INFILTRATION**]

    For Green-Ampt and Modified Green-Ampt Infiltration

    Purpose:
        Supplies infiltration parameters for each subcatchment.
        Rainfall lost to infiltration only occurs over the pervious sub-area of a subcatchment.

    Formats:
        ::

            Subcat Psi Ksat IMD

    Args:
        Subcatch (str): subcatchment name. ``Subcat``
        Psi (float): soil capillary suction (in or mm).
        Ksat (float): soil saturated hydraulic conductivity (in/hr or mm/hr).
        IMD (float): initial soil moisture deficit (volume of voids / total volume).

    Attributes:
        Subcatch (str): subcatchment name. ``Subcat``
        Psi (float): soil capillary suction (in or mm).
        Ksat (float): soil saturated hydraulic conductivity (in/hr or mm/hr).
        IMD (float): initial soil moisture deficit (volume of voids / total volume).
    """

    def __init__(self, Subcatch, Psi, Ksat, IMD, kind=None):
        Infiltration.__init__(self, Subcatch)
        self.Psi = float(Psi)
        self.Ksat = float(Ksat)
        self.IMD = float(IMD)
        self.kind = NaN


class InfiltrationCurveNumber(Infiltration):
    """
    Section: [**INFILTRATION**]

    For Curve-Number Infiltration:

    Purpose:
        Supplies infiltration parameters for each subcatchment.
        Rainfall lost to infiltration only occurs over the pervious sub-area of a subcatchment.

    Formats:
        ::

            Subcat CurveNo Ksat DryTime

    Args:
        Subcatch (str): subcatchment name. ``Subcat``
        CurveNo: SCS Curve Number.
        Ksat (float): soil saturated hydraulic conductivity (in/hr or mm/hr)
            (This property has been deprecated and is no longer used.)
        DryTime (float): time it takes for fully saturated soil to dry (days).

    Attributes:
        Subcatch (str): subcatchment name. ``Subcat``
        CurveNo: SCS Curve Number.
        Ksat (float): soil saturated hydraulic conductivity (in/hr or mm/hr)
            (This property has been deprecated and is no longer used.)
        DryTime (float): time it takes for fully saturated soil to dry (days).
    """

    def __init__(self, Subcatch, CurveNo, Ksat, DryTime, kind=None):
        Infiltration.__init__(self, Subcatch)
        self.CurveNo = CurveNo
        self.Ksat = float(Ksat)
        self.DryTime = float(DryTime)
        self.kind = NaN


INFILTRATION_DICT = {
    'HORTON'             : InfiltrationHorton,
    'MODIFIED_HORTON'    : InfiltrationHorton,
    'GREEN_AMPT'         : InfiltrationGreenAmpt,
    'MODIFIED_GREEN_AMPT': InfiltrationGreenAmpt,
    'CURVE_NUMBER'       : InfiltrationCurveNumber
}


class Polygon(BaseSectionObject):
    """
    Section: [**POLYGONS**]

    Purpose:
        Assigns X,Y coordinates to vertex points of polygons that define a subcatchment boundary.

    Format:
        ::

            Link Xcoord Ycoord

    Remarks:
        Include a separate line for each vertex of the subcatchment polygon, ordered in a
        consistent clockwise or counter-clockwise sequence.

    Args:
        Subcatch (str): name of subcatchment. ``Subcat``
        polygon (list[list[float,float]]): coordinate of the polygon relative to origin in lower left of map.
            - Xcoord: horizontal coordinate of vertex
            - Ycoord: vertical coordinate of vertex

    Attributes:
        Subcatch (str): name of subcatchment. ``Subcat``
        polygon (list[list[float,float]]): coordinate of the polygon relative to origin in lower left of map.
            - Xcoord: horizontal coordinate of vertex
            - Ycoord: vertical coordinate of vertex
    """
    _identifier = IDENTIFIERS.Subcatch
    _table_inp_export = False
    _section_label = POLYGONS

    def __init__(self, Subcatch, polygon):
        self.Subcatch = str(Subcatch)
        self.polygon = polygon

    @classmethod
    def _convert_lines(cls, multi_line_args):
        polygon = list()
        last = None

        for line in multi_line_args:
            Subcatch, x, y = line
            x = float(x)
            y = float(y)
            if Subcatch == last:
                polygon.append([x, y])
            else:
                if last is not None:
                    yield cls(last, polygon)
                last = Subcatch
                polygon = [[x, y]]
        # last
        if last is not None:
            yield cls(last, polygon)

    @property
    def frame(self):
        return DataFrame.from_records(self.polygon, columns=['x', 'y'])

    def to_inp_line(self):
        return '\n'.join([f'{self.Subcatch}  {x:{GIS_FLOAT_FORMAT}} {y:{GIS_FLOAT_FORMAT}}' for x, y in self.polygon])


class Loading(BaseSectionObject):
    """
    Section: [**LOADINGS**]

    Purpose:
        Specifies the pollutant buildup that exists on each subcatchment at the start of a simulation.

    Format:
        ::

            Subcat Pollut InitBuildup Pollut InitBuildup ...

    Format-PCSWMM:
        ``Subcatchment Pollutant Buildup``

    Remarks:
        More than one pair of pollutant - buildup values can be entered per line. If more than
        one line is needed, then the subcatchment name must still be entered first on the
        succeeding lines.

        If an initial buildup is not specified for a pollutant, then its initial buildup is computed
        by applying the DRY_DAYS option (specified in the [OPTIONS] section) to the
        pollutant’s buildup function for each land use in the subcatchment.

    Args:
        Subcatch (str): name of a subcatchment.
        pollutant_buildup (list[list[str, float]]): tuple of

            - Pollut: name of a pollutant.
            - InitBuildup: initial buildup of pollutant (lbs/acre or kg/hectare).

    Attributes:
        Subcatch (str): name of a subcatchment.
        pollutant_buildup (list[list[str, float]]): tuple of

            - Pollut: name of a pollutant.
            - InitBuildup: initial buildup of pollutant (lbs/acre or kg/hectare).

    """
    _identifier = IDENTIFIERS.Subcatch
    _table_inp_export = False
    _section_label = LOADINGS

    def __init__(self, Subcatch, pollutant_buildup_dict=None):
        self.Subcatch = str(Subcatch)
        self.pollutant_buildup_dict = dict()
        if pollutant_buildup_dict:
            self.pollutant_buildup_dict = pollutant_buildup_dict

    def _add_buildup(self, pollutant, buildup):
        self.pollutant_buildup_dict[pollutant] = float(buildup)

    @classmethod
    def _convert_lines(cls, multi_line_args):
        last = None
        for Subcatch, *line in multi_line_args:

            if last is None:
                # first line of section
                last = cls(Subcatch)

            elif last.Subcatch != Subcatch:
                # new Coverage
                yield last
                last = cls(Subcatch)

            # Coverage definitions
            remains = iter(line)
            for pollutant in remains:
                buildup = next(remains)
                last._add_buildup(pollutant, buildup)

        # last
        if last is not None:
            yield last

    @property
    def frame(self):
        return DataFrame.from_dict(self.pollutant_buildup_dict, columns=['pollutant', 'initial buildup'])

    def to_inp_line(self):
        return '\n'.join(['{}  {} {}'.format(self.Subcatch, p, b) for p, b in self.pollutant_buildup_dict.items()])


class Coverage(BaseSectionObject):
    """
    Section: [**COVERAGES**]

    Purpose:
        Specifies the percentage of a subcatchment’s area that is covered by each category of land use.

    Format:
        ::

            Subcat Landuse Percent Landuse Percent . . .

    Args:
        Subcatch (str):
            subcatchment name.

        land_use_dict (dict):
            key: Landuse (str): land use name.
            value: Percent (float): percent of subcatchment area.

    Remarks:
        More than one pair of land use - percentage values can be entered per line. If more
        than one line is needed, then the subcatchment name must still be entered first on
        the succeeding lines.

        If a land use does not pertain to a subcatchment, then it does not have to be entered.

        If no land uses are associated with a subcatchment then no contaminants will appear
        in the runoff from the subcatchment.
    """
    _identifier = IDENTIFIERS.Subcatch
    _section_label = COVERAGES

    def __init__(self, Subcatch, land_use_dict=None):
        self.Subcatch = str(Subcatch)
        self.land_use_dict = dict()
        if land_use_dict:
            self.land_use_dict = land_use_dict

    def _add_land_use(self, land_use, percent):
        self.land_use_dict[land_use] = float(percent)

    @classmethod
    def _convert_lines(cls, multi_line_args):
        last = None
        for Subcatch, *line in multi_line_args:

            if last is None:
                # first line of section
                last = cls(Subcatch)

            elif last.Subcatch != Subcatch:
                # new Coverage
                yield last
                last = cls(Subcatch)

            # Coverage definitions
            remains = iter(line)
            for land_use in remains:
                percent = next(remains)
                last._add_land_use(land_use, percent)

        # last
        if last is not None:
            yield last

    @property
    def frame(self):
        return DataFrame.from_dict(self.land_use_dict, columns=['land_use', 'percent'])

    def to_inp_line(self):
        return '\n'.join(['{}  {} {}'.format(self.Subcatch, p, b) for p, b in self.land_use_dict.items()])


class GroundwaterFlow(BaseSectionObject):
    """
    Section: [**GWF**]

    Purpose:
        Defines custom groundwater flow equations for specific subcatchments.

    Format:
        ::

            Subcat LATERAL/DEEP Expr

    Args:
        Subcatch (str): subcatchment name.
        expression (str): math formula expressing the rate of groundwater flow (in cfs per acre or cms per hectare for lateral flow or in/hr or mm/hr for deep flow) as a function of the following variables:

            - ``Hgw`` (for height of the groundwater table)
            - ``Hsw`` (for height of the surface water)
            - ``Hcb`` (for height of the channel bottom)
            - ``Hgs`` (for height of ground surface) where all heights are relative to the aquifer bottom and have units of either feet or meters;
            - ``Ks`` (for saturated hydraulic conductivity in in/hr or mm/hr)
            - ``K`` (for unsaturated hydraulic conductivity in in/hr or mm/hr)
            - ``Theta`` (for moisture content of unsaturated zone)
            - ``Phi`` (for aquifer soil porosity)
            - ``Fi`` (for infiltration rate from the ground surface in in/hr or mm/hr)
            - ``Fu`` (for percolation rate from the upper unsaturated zone in in/hr or mm/hr)
            - ``A`` (for subcatchment area in acres or hectares)

    Remarks:
        Use ``LATERAL`` to designate an expression for lateral groundwater flow (to a node of
        the conveyance network) and ``DEEP`` for vertical loss to deep groundwater.

        See the [``TREATMENT``] section for a list of built-in math functions that can be used in
        ``Expr``. In particular, the ``STEP(x)`` function is 1 when ``x > 0`` and is 0 otherwise.

    Examples:
        ::

            ;Two-stage linear reservoir for lateral flow
            Subcatch1 LATERAL 0.001*Hgw + 0.05*(Hgw–5)*STEP(Hgw–5)

            ;Constant seepage rate to deep aquifer
            Subactch1 DEEP 0.002
    """
    _identifier = (IDENTIFIERS.Subcatch, 'kind')
    _section_label = GWF

    class TYPES:
        LATERAL = 'LATERAL'
        DEEP = 'DEEP'

    def __init__(self, Subcatch, kind, expression, *expression_):
        self.Subcatch = str(Subcatch)
        self.kind = kind
        self.expression = expression + ' '.join(expression_)


class Groundwater(BaseSectionObject):
    """
    Section: [**GROUNDWATER**]

    Purpose:
        Supplies parameters that determine the rate of groundwater flow between the aquifer
        underneath a subcatchment and a node of the conveyance system.

    Format:
        ::

            Subcat Aquifer Node Esurf A1 B1 A2 B2 A3 Dsw (Egwt Ebot Egw Umc)

    Args:
        Subcat (float): subcatchment name.
        Aquifer (float): name of groundwater aquifer underneath the subcatchment.
        Node (float): name of node in conveyance system exchanging groundwater with aquifer.
        Esurf (float): surface elevation of subcatchment (ft or m).
        A1 (float): groundwater flow coefficient (see below).
        B1 (float): groundwater flow exponent (see below).
        A2 (float): surface water flow coefficient (see below).
        B2 (float): surface water flow exponent (see below).
        A3 (float): surface water – groundwater interaction coefficient (see below).
        Dsw (float): fixed depth of surface water at receiving node (ft or m)
                    (set to zero if surface water depth will vary as computed by flow routing).
        Egwt (float): threshold groundwater table elevation which must be reached before any flow occurs (ft or m).
                    Leave blank (or enter \\*) to use the elevation of the receiving node's invert.
        Ebot (float): elevation of the bottom of the aquifer (ft or m).
        Egw (float): groundwater table elevation at the start of the simulation (ft or m).
        Umc (float): unsaturated zone moisture content at start of simulation (volumetric fraction).


    Remarks:
        The following optional parameters can be used to override the values supplied for the subcatchment’s aquifer:
            - Ebot
            - Egw
            - Umc

        The flow coefficients are used in the following equation that determines the lateral groundwater
        flow rate based on groundwater and surface water elevations:

            Q_L = A1 * (H_gw – H_cb ) ^ B1 – A2 * (H_sw – H_cb ) ^ B2 + A3 * H_gw * H_sw

        where:
            - Q_L = lateral groundwater flow (cfs per acre or cms per hectare),
            - H_gw = height of saturated zone above bottom of aquifer (ft or m),
            - H_sw = height of surface water at receiving node above aquifer bottom (ft or m),
            - H_cb = height of channel bottom above aquifer bottom (ft or m).
    """
    _identifier = (IDENTIFIERS.Subcatch, 'Aquifer', IDENTIFIERS.Node)
    _section_label = GROUNDWATER

    def __init__(self, Subcatch, Aquifer, Node, Esurf, A1, B1, A2, B2, A3, Dsw, Egwt=NaN, Ebot=NaN, Egw=NaN, Umc=NaN):
        self.Subcatch = str(Subcatch)
        self.Aquifer = str(Aquifer)
        self.Node = str(Node)
        self.Esurf = float(Esurf)
        self.A1 = float(A1)
        self.B1 = float(B1)
        self.A2 = float(A2)
        self.B2 = float(B2)
        self.A3 = float(A3)
        self.Dsw = float(Dsw)
        self.Egwt = Egwt
        self.Ebot = Ebot
        self.Egw = Egw
        self.Umc = Umc



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

import pandas as pd

"""SWMM5 User Manual: Section 11.4 Climate Files
A user-prepared climate file where each line contains a recording station name, the year, month, day, 
maximum temperature, minimum temperature, and optionally, evaporation rate, and wind speed. 
If no data are available for any of these items on a given date, then an asterisk should be entered as its value

For a user-prepared climate file, the data must be in the same units as the project being
analyzed. 

For US units, 
    temperature is in degrees F, 
    evaporation is in inches/day, and
    wind speed is in miles/hour. 

For metric units, 
    temperature is in degrees C, 
    evaporation is in mm/day, and 
    wind speed is in km/hour.
"""


class COLUMNS:
    STATION = 'station name'
    MAX_TEMPERATURE = 'maximum temperature'
    MIN_TEMPERATURE = 'minimum temperature'
    EVAPORATION = 'evaporation rate'
    WIND_SPEED = 'wind speed'

    YEAR = 'year'
    MONTH = 'month'
    DAY = 'day'

    FOR_FRAME = [STATION, MAX_TEMPERATURE, MIN_TEMPERATURE, EVAPORATION, WIND_SPEED]
    FOR_FILE = [STATION, YEAR, MONTH, DAY, MAX_TEMPERATURE, MIN_TEMPERATURE, EVAPORATION, WIND_SPEED]


def read_climate_dat_file(filename):
    """
    read climate .dat-file

    Args:
        filename (str): path and filename of the file to be read

    Returns:
        pandas.DataFrame: climate dataframe
    """
    df = pd.read_csv(filename, sep=' ', decimal='.', parse_dates={'date': [1, 2, 3]},
                     keep_date_col=False, header=None, index_col=['date'], na_values='*')
    df.columns = COLUMNS.FOR_FRAME
    return df


def write_climate_dat_file(df, filename):
    """
    write climate .dat-file for SWMM

    Args:
        df (pandas.DataFrame): climate dataframe:
        filename (str): path and filename for the new file
    """
    df[COLUMNS.YEAR] = df.index.year
    df[COLUMNS.MONTH] = df.index.strftime('%m')
    df[COLUMNS.DAY] = df.index.strftime('%d')

    df[COLUMNS.FOR_FILE].to_csv(filename, sep=' ', header=None, index=None, na_rep='*')



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

import pandas as pd


def _following_numbers_bool_base(series):
    return (series.diff() != 0) | (series.diff(-1) != 0)


def following_numbers_bool(series):
    """mark first and last value of following duplicates BUT no NaN values"""
    return _following_numbers_bool_base(series) & ~series.isna()


def following_nans_bool(series):
    """marks first and last NaN value of following NaN duplicates AND the value before and after NaN values"""
    return _following_numbers_bool_base(series.isna().astype(int))


def remove_following_zeros(series):
    """
    removes all following zeros (null) and NaN
    to reduce the data for plots etc.
    only tested with int/float/NaN

    Args:
        series (pandas.Series): time series

    Returns:
        pandas.Series: reduced time series
    """
    bool_series = ((_following_numbers_bool_base(series) | (series != 0)) & ~series.isna()) | following_nans_bool(series)
    return series.loc[bool_series].copy()


def remove_following_duplicates(data):
    """
    removes all following duplicates
    to reduce the data for plots etc.
    only tested with int/float/NaN
    only first and last following occurrence will stay

    OR:
    remove rows where the value is the sam as in the previous row, BUT keep the last value

    Args:
        data (pandas.DataFrame | pandas.Series):

    Returns:
        pandas.DataFrame | pandas.Series:
    """
    if isinstance(data, pd.DataFrame):
        bool_series = pd.Series(index=data.index, data=False)
        for column in data:
            series = data[column]
            bool_series |= following_numbers_bool(series) | following_nans_bool(series)
    # ________________________________
    elif isinstance(data, pd.Series):
        bool_series = following_numbers_bool(data) | following_nans_bool(data)
    # ________________________________
    else:
        raise NotImplementedError('"{}" not implemented'.format(type(data)))
    # ________________________________
    return data.loc[bool_series].copy()


from ..section_labels import *
from .collection import links_dict, nodes_dict
from ..misc.curve_simplification import ramer_douglas
from ..section_types import SECTION_TYPES
from ..sections import Control


def reduce_curves(inp):
    """
    remove unused curves

    only keep used CURVES from the sections [STORAGE, OUTLETS, PUMPS and XSECTIONS]

    Args:
        inp (swmm_api.SwmmInput): inp-file data

    .. Important::
        works inplace
    """
    if CURVES not in inp:
        return inp
    used_curves = set()
    for section in [STORAGE, OUTLETS, PUMPS, XSECTIONS]:
        if section in inp:
            for name in inp[section]:
                if isinstance(inp[section][name].Curve, str):
                    used_curves.add(inp[section][name].Curve)

    inp[CURVES] = inp[CURVES].slice_section(used_curves)


def reduce_pattern(inp):
    used_pattern = set()
    if EVAPORATION in inp:
        #  optional monthly time pattern of multipliers for infiltration recovery rates during dry periods
        if 'RECOVERY' in inp[EVAPORATION]:
            used_pattern.add(inp[EVAPORATION]['RECOVERY'])

    if AQUIFERS in inp:
        #  optional monthly time pattern used to adjust the upper zone evaporation fraction
        used_pattern |= set(inp[AQUIFERS].frame['Epat'].dropna().values)

    if INFLOWS in inp:
        #  optional time pattern used to adjust the baseline value on a periodic basis
        used_pattern |= set(inp[INFLOWS].frame['Pattern'].dropna().values)

    if DWF in inp:
        for i in range(1, 5):
            used_pattern |= set(inp[DWF].frame[f'pattern{i}'].dropna().values)

    if PATTERNS in inp:
        inp[PATTERNS] = inp[PATTERNS].slice_section(used_pattern)


def reduce_controls(inp):
    """
    remove unused controls

    only keep used CONTROLS the sections [CONDUIT, ORIFICE, WEIR, OUTLET / NODE, LINK, CONDUIT, PUMP, ORIFICE, WEIR, OUTLET]

    if unavailable object in condition: remove whole rule
    if unavailable object in action: remove only this action

    Args:
        inp (swmm_api.SwmmInput): inp-file data

    .. Important::
        works inplace
    """
    if CONTROLS not in inp:
        return

    # c0 = Control.create_section(inp.CONTROLS.to_inp_lines())

    links = links_dict(inp)
    nodes = nodes_dict(inp)

    for label in list(inp.CONTROLS.keys()):
        control = inp.CONTROLS[label]
        # if unavailable object in condition: remove whole rule
        for condition in control.conditions:
            if condition.kind + 'S' in inp:
                # CONDUIT PUMP ORIFICE WEIR OUTLET
                if condition.label not in inp[condition.kind + 'S']:
                    # pass  # delete
                    del inp.CONTROLS[label]
                    continue
            elif condition.kind == Control.OBJECTS.NODE:
                if condition.label not in nodes:
                    # pass  # delete
                    del inp.CONTROLS[label]
                    continue
            elif condition.kind == Control.OBJECTS.LINK:
                if condition.label not in links:
                    # pass  # delete
                    del inp.CONTROLS[label]
                    continue

        # if unavailable object in action: remove only this action
        for action in list(control.actions):
            i = control.actions.index(action)
            if action.kind + 'S' in inp:
                # CONDUIT PUMP ORIFICE WEIR OUTLET
                if action.label not in inp[action.kind + 'S']:
                    # pass  # delete
                    inp.CONTROLS[label].actions.pop(i)
            elif action.kind + 'S' in SECTION_TYPES and action.kind + 'S' not in inp:
                inp.CONTROLS[label].actions.pop(i)
            elif action.kind == Control.OBJECTS.NODE:
                if action.label not in nodes:
                    # pass  # delete
                    inp.CONTROLS[label].actions.pop(i)
            elif action.kind == Control.OBJECTS.LINK:
                if action.label not in links:
                    # pass  # delete
                    inp.CONTROLS[label].actions.pop(i)

        # if no actions left
        if not control.actions:
            del inp.CONTROLS[label]


def simplify_curves(curve_section, dist=0.001):
    """
    simplify curves with the algorithm by Ramer and Douglas

    Args:
        curve_section (InpSection[Curve]): old section
        dist (float): maximum Ramer-Douglas distance

    Returns:
        InpSection[Curve]: new section
    """
    # new = Curve.create_section()
    # for label, curve in curve_section.items():
    #     new[label] = Curve(curve.Name, curve.Type, points=ramer_douglas(curve_section[label].points, dist=dist))
    # return new
    for curve in curve_section.values():  # type: Curve
        curve.points = ramer_douglas(curve.points, dist=dist)
    return curve_section


def reduce_raingages(inp):
    """
    get used RAINGAGES from SUBCATCHMENTS and keep only used raingages in the section

    Args:
        inp (SwmmInput):  inp-file data

    Returns:
        SwmmInput: inp-file data with filtered RAINGAGES section
    """
    if SUBCATCHMENTS not in inp or RAINGAGES not in inp:
        return inp
    needed_raingages = {inp[SUBCATCHMENTS][s].RainGage for s in inp[SUBCATCHMENTS]}
    inp[RAINGAGES] = inp[RAINGAGES].slice_section(needed_raingages)


def remove_empty_sections(inp):
    """
    remove empty inp-file data sections

    Args:
        inp (SwmmInput): inp-file data

    Returns:
        SwmmInput: cleaned inp-file data
    """
    for section in list(inp.keys()):
        if not inp[section]:
            del inp[section]



import os
import re
import warnings

from .helpers import (section_to_string, CustomDict, convert_section, InpSection,
                      InpSectionGeneric, SECTION_ORDER_DEFAULT, check_order, SECTIONS_ORDER_MP, head_to_str,
                      iter_section_lines, SwmmInputWarning, )
from .section_types import SECTION_TYPES
from .section_labels import *
from .sections import *
from .sections.subcatch import INFILTRATION_DICT


class SwmmInput(CustomDict):
    """
    overall class for an input file

    child class of dict

    just used for the copy function and to identify ``.inp``-file data
    """

    def __init__(self, *args, custom_section_handler=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._converter = SECTION_TYPES.copy()

        if custom_section_handler is not None:
            self._converter.update(custom_section_handler)

        # only when reading a new file
        self._original_section_order = SECTIONS_ORDER_MP

    def copy(self):
        new = type(self)()
        for key in self:
            if hasattr(self._data[key], 'copy'):
                new._data[key] = self._data[key].copy()
            else:
                new._data[key] = self._data[key]
        return new

    def update(self, d=None, **kwargs):
        for sec in d:
            if sec not in self:
                self[sec] = d[sec]
            else:
                if isinstance(self[sec], str):
                    pass
                else:
                    self[sec].update(d[sec])

    @classmethod
    def read_file(cls, filename, custom_converter=None, force_ignore_case=False, encoding='iso-8859-1'):
        """
        read ``.inp``-file and convert the sections in pythonic objects

        the sections will be converted when used

        Args:
            filename (str): path/filename to .inp file
            custom_converter (dict): dictionary of {section: converter/section_type} Default: :py:const:`SECTION_TYPES`
            force_ignore_case (bool): SWMM is case-insensitive but python is case-sensitive -> set True to ignore case
                                        all text/labels will be set to uppercase
            encoding (str): encoding of the inp text file

        Returns:
            SwmmInput: dict-like data of the sections in the ``.inp``-file
        """
        inp = cls(custom_section_handler=custom_converter)
        if os.path.isfile(filename) or filename.endswith('.inp'):
            with open(filename, 'r', encoding=encoding) as inp_file:
                txt = inp_file.read()
        else:
            txt = filename

        # __________________________________
        if force_ignore_case:
            txt = txt.upper()

        # __________________________________
        inp._original_section_order = []
        for head, lines in zip(re.findall(r"\[(\w+)\]", txt),
                               re.split(r"\[\w+\]", txt)[1:]):
            inp._data[head.upper()] = lines.strip()
            inp._original_section_order.append(head.upper())

        # ----------------
        # if order in inp follows default SWMM GUI order
        #   set full/complete order list of SWMM GUI
        #   to use the right order for additional created setions
        if check_order(inp, SECTION_ORDER_DEFAULT):
            inp._original_section_order = SECTION_ORDER_DEFAULT

        inp.set_default_infiltration_from_options()
        return inp

    def force_convert_all(self):
        _ = [self[sec] for sec in self]

    def __getitem__(self, key):
        # if section not in inp-data, create an empty section
        if key not in self:
            self._data[key] = self._converter[key].create_section()

        # if section is a string (raw string from the .inp-file) convert section first
        if isinstance(self._data[key], str):
            self._data[key] = convert_section(key, self._data[key], self._converter)

        return self._data.__getitem__(key)

    def __setitem__(self, key, item):
        super().__setitem__(key, item)
        # super().__setattr__(key, item)
        # self._data.__setitem__(key, item)
        # if key == OPTIONS:
        #     self.set_default_infiltration_from_options()
        if hasattr(self._data[key], 'set_parent_inp'):
            self._data[key].set_parent_inp(self)

    def __delattr__(self, item):
        """delete section"""
        if item in SECTION_TYPES:
            # print('delete section')
            del self._data[item]
        else:
            # print('delete attribute')
            super().__delattr__(item)

    def set_default_infiltration_from_options(self):
        if OPTIONS in self \
                and 'INFILTRATION' in self[OPTIONS] \
                and isinstance(self[OPTIONS], (dict, OptionSection, InpSectionGeneric)):
            self.set_infiltration_method(INFILTRATION_DICT.get(self[OPTIONS]['INFILTRATION']))

    def set_infiltration_method(self, infiltration_class):
        self._converter[INFILTRATION] = infiltration_class

    def get_section_headers(self, custom_sections_order=None):
        if custom_sections_order is None:
            custom_sections_order = self._original_section_order

        def _sort_by(key):
            if key in custom_sections_order:
                return custom_sections_order.index(key)
            else:
                return len(custom_sections_order)

        return sorted(self.keys(), key=_sort_by)

    def to_string(self, fast=True, custom_sections_order=None, sort_objects_alphabetical=False):
        """
        create the string of a new ``.inp``-file

        Args:
            fast (bool): don't use any formatting else format as table
            custom_sections_order (list[str]): list of section names to preset the order of the section in the
                created inp-file | default: order of the read inp-file + default order of the SWMM GUI
            sort_objects_alphabetical (bool): if objects in a section should be sorted alphabetical |
                default: use order of the read inp-file and append new objects

        Returns:
            str: string of input file text
        """
        f = ''
        for head in self.get_section_headers(custom_sections_order):
            f += head_to_str(head)
            f += section_to_string(self._data[head], fast=fast, sort_objects_alphabetical=sort_objects_alphabetical)
        return f

    def write_file(self, filename, fast=True, encoding='iso-8859-1', custom_sections_order=None,
                   sort_objects_alphabetical=False, per_line=False):
        """
        create/write a new ``.inp``-file

        Args:
            filename (str): path/filename of created ``.inp``-file
            fast (bool): don't use any formatting else format as table
            encoding (str): define encoding for resulting inp-file
            custom_sections_order (list[str]): list of section names to preset the order of the section in the created
                inp-file | default: order of the read inp-file + default order of the SWMM GUI
            sort_objects_alphabetical (bool): if objects in a section should be sorted alphabetical |
                default: use order of the read inp-file and append new objects
            per_line (bool): weather to write the data line per line (=True) or section per section (=False) into the
                file. line per line has an advantage for big files (> 1 GB) and uses less memory (RAM).
        """
        with open(filename, 'w', encoding=encoding) as f:
            for head in self.get_section_headers(custom_sections_order):
                f.write(head_to_str(head))
                if per_line:
                    for line in iter_section_lines(self._data[head],
                                                   sort_objects_alphabetical=sort_objects_alphabetical):
                        f.write(line + '\n')
                else:
                    f.write(section_to_string(self._data[head], fast=fast,
                                              sort_objects_alphabetical=sort_objects_alphabetical))
        return filename

    def print_string(self, custom_sections_order=None):
        for head in self.get_section_headers(custom_sections_order):
            print(head_to_str(head))
            for line in iter_section_lines(self._data[head], sort_objects_alphabetical=False):
                print(line)

    def check_for_section(self, obj):
        """
        check if a section is in the inp-data, and create it if not present

        Args:
            obj (BaseSectionObject or InpSectionGeneric):

        Returns:
            InpSection or InpSectionGeneric: section of inp
        """
        if obj._section_label not in self:
            self[obj._section_label] = obj.create_section()
        return self[obj._section_label]

    def add_new_section(self, section):
        """
        add new section to the inp-data

        Args:
            section (InpSection, InpSectionGeneric):

        .. Important::
            works inplace
        """
        if section._label not in self:
            self[section._label] = section
        else:
            warnings.warn(f'Section [{section._label}] not empty!', SwmmInputWarning)

    def add_obj(self, obj):
        """
        add object to respective section

        Args:
            obj (BaseSectionObject):new object
        """
        self.check_for_section(obj)
        self[obj._section_label].add_obj(obj)

    def add_multiple(self, *items):
        """
        add multiple objects to respective sections

        Args:
            *items (BaseSectionObject): new objects
        """
        for obj in items:
            self.add_obj(obj)

    @property
    def OPTIONS(self):
        """
        OPTIONS Section

        Returns:
            OptionSection: OPTIONS Section
        """
        if OPTIONS in self:
            return self[OPTIONS]

    @property
    def REPORT(self):
        """
        REPORT Section

        Returns:
            ReportSection: REPORT Section
        """
        if REPORT in self:
            return self[REPORT]

    @property
    def EVAPORATION(self):
        """
        EVAPORATION Section

        Returns:
            EvaporationSection: EVAPORATION Section
        """
        if EVAPORATION in self:
            return self[EVAPORATION]

    @property
    def TEMPERATURE(self):
        """
        TEMPERATURE Section

        Returns:
            TemperatureSection: TEMPERATURE Section
        """
        if TEMPERATURE in self:
            return self[TEMPERATURE]

    @property
    def FILES(self):
        """
        FILES Section

        Returns:
            FilesSection: FILES Section
        """
        if FILES in self:
            return self[FILES]

    @property
    def BACKDROP(self):
        """
        BACKDROP Section

        Returns:
            BackdropSection: BACKDROP Section
        """
        if BACKDROP in self:
            return self[BACKDROP]

    @property
    def ADJUSTMENTS(self):
        """
        ADJUSTMENTS Section

        Returns:
            AdjustmentsSection: ADJUSTMENTS Section
        """
        if ADJUSTMENTS in self:
            return self[ADJUSTMENTS]

    # -----
    @property
    def COORDINATES(self):
        """
        COORDINATES section

        Returns:
            dict[str, Coordinate | CoordinateGeo] | InpSection: Coordinates in INP
        """
        if COORDINATES in self:
            return self[COORDINATES]

    @property
    def VERTICES(self):
        """
        VERTICES section

        Returns:
            dict[str, Vertices] | InpSection: Vertices section
        """
        if VERTICES in self:
            return self[VERTICES]

    @property
    def POLYGONS(self):
        """
        POLYGONS section

        Returns:
            dict[str, Polygon] | InpSection: Polygon section
        """
        if POLYGONS in self:
            return self[POLYGONS]

    @property
    def SYMBOLS(self):
        """
        SYMBOLS section

        Returns:
            dict[str, Symbol] | InpSection: Symbol section
        """
        if SYMBOLS in self:
            return self[SYMBOLS]

    @property
    def MAP(self):
        """
        MAP section

        Returns:
            dict[str, MapSection] | InpSection: MapSection section
        """
        if MAP in self:
            return self[MAP]

    @property
    def LABELS(self):
        """
        LABELS section

        Returns:
            dict[str, Label] | InpSection: Label section
        """
        if LABELS in self:
            return self[LABELS]

    @property
    def CONDUITS(self):
        """
        CONDUITS section

        Returns:
            dict[str, Conduit] | InpSection: Conduit section
        """
        if CONDUITS in self:
            return self[CONDUITS]

    @property
    def ORIFICES(self):
        """
        ORIFICES section

        Returns:
            dict[str, Orifice] | InpSection: Orifice section
        """
        if ORIFICES in self:
            return self[ORIFICES]

    @property
    def WEIRS(self):
        """
        WEIRS section

        Returns:
            dict[str, Weir] | InpSection: Weir section
        """
        if WEIRS in self:
            return self[WEIRS]

    @property
    def PUMPS(self):
        """
        PUMPS section

        Returns:
            dict[str, Pump] | InpSection: Pump section
        """
        if PUMPS in self:
            return self[PUMPS]

    @property
    def OUTLETS(self):
        """
        OUTLETS section

        Returns:
            dict[str, Outlet] | InpSection: Outlet section
        """
        if OUTLETS in self:
            return self[OUTLETS]

    @property
    def TRANSECTS(self):
        """
        TRANSECTS section

        Returns:
            dict[str, Transect] | InpSection: Transect section
        """
        if TRANSECTS in self:
            return self[TRANSECTS]

    @property
    def XSECTIONS(self):
        """
        XSECTIONS section

        Returns:
            dict[str, CrossSection] | InpSection: CrossSection section
        """
        if XSECTIONS in self:
            return self[XSECTIONS]

    @property
    def LOSSES(self):
        """
        LOSSES section

        Returns:
            dict[str, Loss] | InpSection: Loss section
        """
        if LOSSES in self:
            return self[LOSSES]

    @property
    def JUNCTIONS(self):
        """
        JUNCTIONS section

        Returns:
            dict[str, Junction] | InpSection: Junction section
        """
        if JUNCTIONS in self:
            return self[JUNCTIONS]

    @property
    def OUTFALLS(self):
        """
        OUTFALLS section

        Returns:
            dict[str, Outfall] | InpSection: Outfall section
        """
        if OUTFALLS in self:
            return self[OUTFALLS]

    @property
    def STORAGE(self):
        """
        STORAGE section

        Returns:
            dict[str, swmm_api.input_file.sections.node.Storage] | InpSection: Storage section
        """
        if STORAGE in self:
            return self[STORAGE]

    @property
    def DWF(self):
        """
        DWF section

        Returns:
            dict[str, DryWeatherFlow] | InpSection: DryWeatherFlow section
        """
        if DWF in self:
            return self[DWF]

    @property
    def INFLOWS(self):
        """
        INFLOWS section

        Returns:
            dict[str, Inflow] | InpSection: Inflow section
        """
        if INFLOWS in self:
            return self[INFLOWS]

    @property
    def RDII(self):
        """
        RDII section

        Returns:
            dict[str, RainfallDependentInfiltrationInflow] | InpSection: RainfallDependentInfiltrationInflow section
        """
        if RDII in self:
            return self[RDII]

    @property
    def TREATMENT(self):
        """
        TREATMENT section

        Returns:
            dict[str, Treatment] | InpSection: Treatment section
        """
        if TREATMENT in self:
            return self[TREATMENT]

    @property
    def SUBCATCHMENTS(self):
        """
        SUBCATCHMENTS section

        Returns:
            dict[str, SubCatchment] | InpSection: SubCatchment section
        """
        if SUBCATCHMENTS in self:
            return self[SUBCATCHMENTS]

    @property
    def SUBAREAS(self):
        """
        SUBAREAS section

        Returns:
            dict[str, SubArea] | InpSection: SubArea section
        """
        if SUBAREAS in self:
            return self[SUBAREAS]

    @property
    def INFILTRATION(self):
        """
        INFILTRATION section

        Returns:
            dict[str, Infiltration] | InpSection: Infiltration section
        """
        if INFILTRATION in self:
            return self[INFILTRATION]

    @property
    def LOADINGS(self):
        """
        LOADINGS section

        Returns:
            dict[str, Loading] | InpSection: Loading section
        """
        if LOADINGS in self:
            return self[LOADINGS]

    @property
    def WASHOFF(self):
        """
        WASHOFF section

        Returns:
            dict[str, WashOff] | InpSection: WashOff section
        """
        if WASHOFF in self:
            return self[WASHOFF]

    @property
    def BUILDUP(self):
        """
        BUILDUP section

        Returns:
            dict[str, BuildUp] | InpSection: BuildUp section
        """
        if BUILDUP in self:
            return self[BUILDUP]

    @property
    def COVERAGES(self):
        """
        COVERAGES section

        Returns:
            dict[str, Coverage] | InpSection: Coverage section
        """
        if COVERAGES in self:
            return self[COVERAGES]

    @property
    def GWF(self):
        """
        GWF section

        Returns:
            dict[str, GroundwaterFlow] | InpSection: GroundwaterFlow section
        """
        if GWF in self:
            return self[GWF]

    @property
    def GROUNDWATER(self):
        """
        GROUNDWATER section

        Returns:
            dict[str, Groundwater] | InpSection: Groundwater section
        """
        if GROUNDWATER in self:
            return self[GROUNDWATER]

    @property
    def RAINGAGES(self):
        """
        RAINGAGES section

        Returns:
            dict[str, RainGage] | InpSection: RainGage section
        """
        if RAINGAGES in self:
            return self[RAINGAGES]

    @property
    def PATTERNS(self):
        """
        PATTERNS section

        Returns:
            dict[str, Pattern] | InpSection: Pattern section
        """
        if PATTERNS in self:
            return self[PATTERNS]

    @property
    def POLLUTANTS(self):
        """
        POLLUTANTS section

        Returns:
            dict[str, Pollutant] | InpSection: Pollutant section
        """
        if POLLUTANTS in self:
            return self[POLLUTANTS]

    @property
    def CONTROLS(self):
        """
        CONTROLS section

        Returns:
            dict[str, Control] | InpSection: Control section
        """
        if CONTROLS in self:
            return self[CONTROLS]

    @property
    def CURVES(self):
        """
        CURVES section

        Returns:
            dict[str, Curve] | InpSection: Curve section
        """
        if CURVES in self:
            return self[CURVES]

    @property
    def TIMESERIES(self):
        """
        TIMESERIES section

        Returns:
            dict[str, Timeseries] | InpSection: Timeseries section
        """
        if TIMESERIES in self:
            return self[TIMESERIES]

    @property
    def TAGS(self):
        """
        TAGS section

        Returns:
            dict[str, Tag] | InpSection: Tag section
        """
        if TAGS in self:
            return self[TAGS]

    @property
    def HYDROGRAPHS(self):
        """
        HYDROGRAPHS section

        Returns:
            dict[str, Hydrograph] | InpSection: Hydrograph section
        """
        if HYDROGRAPHS in self:
            return self[HYDROGRAPHS]

    @property
    def LANDUSES(self):
        """
        LANDUSES section

        Returns:
            dict[str, LandUse] | InpSection: LandUse section
        """
        if LANDUSES in self:
            return self[LANDUSES]

    @property
    def AQUIFERS(self):
        """
        AQUIFERS section

        Returns:
            dict[str, Aquifer] | InpSection: Aquifer section
        """
        if AQUIFERS in self:
            return self[AQUIFERS]

    @property
    def SNOWPACKS(self):
        """
        SNOWPACKS section

        Returns:
            dict[str, SnowPack] | InpSection: SnowPack section
        """
        if SNOWPACKS in self:
            return self[SNOWPACKS]

    @property
    def LID_CONTROLS(self):
        """
        LID_CONTROLS section

        Returns:
            dict[str, LIDControl] | InpSection: LIDControl section
        """
        if LID_CONTROLS in self:
            return self[LID_CONTROLS]

    @property
    def LID_USAGE(self):
        """
        LID_USAGE section

        Returns:
            dict[str, LIDUsage] | InpSection: LIDUsage section
        """
        if LID_USAGE in self:
            return self[LID_USAGE]


def read_inp_file(filename, custom_converter=None, force_ignore_case=False, encoding='iso-8859-1'):
    """
    read ``.inp``-file and convert the sections in pythonic objects

    the sections will be converted when used

    Args:
        filename (str): path/filename to .inp file
        custom_converter (dict): dictionary of {section: converter/section_type} Default: :py:const:`SECTION_TYPES`
        force_ignore_case (bool): SWMM is case-insensitive but python is case-sensitive -> set True to ignore case
        encoding (str): encoding of the inp text file

    Returns:
        SwmmInput: dict-like data of the sections in the ``.inp``-file
    """
    return SwmmInput.read_file(filename, custom_converter=custom_converter, force_ignore_case=force_ignore_case,
                               encoding=encoding)



class IDENTIFIERS:
    """predefined to identify the objects in an input file section and also used in the tags section"""
    Name = 'Name'

    Node = 'Node'
    Subcatch = 'Subcatch'
    Link = 'Link'

    Gage = 'Gage'

    Constituent = 'Constituent'

    Pollutant = 'pollutant'

    Landuse = 'landuse'



from .reduce_unneeded import reduce_controls, reduce_curves, reduce_raingages, remove_empty_sections, reduce_pattern
from ..section_labels import *
from ..section_lists import LINK_SECTIONS, NODE_SECTIONS
from ..sections import Tag
from ..sections._identifiers import IDENTIFIERS


def filter_tags(inp):
    pass


def filter_nodes(inp, final_nodes):
    """
     filter nodes in the network

    Args:
        inp (SwmmInput): inp-file data
        final_nodes (list | set):

    Returns:
        SwmmInput: new inp-file data
    """
    for section in NODE_SECTIONS + [COORDINATES]:
        if section in inp:
            inp[section] = inp[section].slice_section(final_nodes)

    # __________________________________________
    for section in [INFLOWS, DWF]:
        if section in inp:
            inp[section] = inp[section].slice_section(final_nodes, by=IDENTIFIERS.Node)

    # __________________________________________
    if TAGS in inp:
        new = inp[TAGS].create_new_empty()
        new.add_multiple(*inp[TAGS].filter_keys([Tag.TYPES.Subcatch, Tag.TYPES.Link], by='kind'))
        new.add_multiple(*inp[TAGS].filter_keys(((Tag.TYPES.Node, k) for k in final_nodes)))
        inp[TAGS] = new

    # __________________________________________
    remove_empty_sections(inp)
    return inp


def filter_links_within_nodes(inp, final_nodes):
    """
    filter links by nodes in the network

    Args:
        inp (SwmmInput): inp-file data
        final_nodes (list | set):

    Returns:
        SwmmInput: new inp-file data
    """
    final_links = set()
    for section in LINK_SECTIONS:
        if section in inp:
            inp[section] = inp[section].slice_section(final_nodes, by=['FromNode', 'ToNode'])
            final_links |= set(inp[section].keys())

    # __________________________________________
    inp = _filter_link_components(inp, final_links)
    # __________________________________________
    remove_empty_sections(inp)
    return inp


def filter_links(inp, final_links):
    """
    filter links by nodes in the network

    Args:
        inp (SwmmInput): inp-file data
        final_links (list | set):

    Returns:
        SwmmInput: new inp-file data
    """
    for section in LINK_SECTIONS:
        if section in inp:
            inp[section] = inp[section].slice_section(final_links)

    # __________________________________________
    inp = _filter_link_components(inp, final_links)
    # __________________________________________
    remove_empty_sections(inp)
    return inp


def _filter_link_components(inp, final_links):
    for section in [XSECTIONS, LOSSES, VERTICES]:
        if section in inp:
            inp[section] = inp[section].slice_section(final_links)

    # __________________________________________
    if TAGS in inp:
        new = inp[TAGS].create_new_empty()
        new.add_multiple(*inp[TAGS].filter_keys([Tag.TYPES.Subcatch, Tag.TYPES.Node], by='kind'))
        new.add_multiple(*inp[TAGS].filter_keys(((Tag.TYPES.Link, k) for k in final_links)))
        inp[TAGS] = new
        # inp[TAGS] = inp[TAGS].slice_section(((Tag.TYPES.Link, k) for k in final_links))

    # __________________________________________
    remove_empty_sections(inp)
    return inp


def filter_subcatchments(inp, final_nodes):
    """
    filter subcatchments by nodes in the network

    Args:
        inp (SwmmInput): inp-file data
        final_nodes (list | set):

    Returns:
        SwmmInput: new inp-file data
    """
    if SUBCATCHMENTS in inp:
        sub_orig = inp[SUBCATCHMENTS].copy()
        # all with an outlet to final_nodes
        inp[SUBCATCHMENTS] = inp[SUBCATCHMENTS].slice_section(final_nodes, by='Outlet')
        # all with an outlet to an subcatchment
        inp[SUBCATCHMENTS].update(sub_orig.slice_section(inp[SUBCATCHMENTS].keys(), by='Outlet'))

        # __________________________________________
        for section in [SUBAREAS, INFILTRATION, POLYGONS]:
            if section in inp:
                inp[section] = inp[section].slice_section(inp[SUBCATCHMENTS])

        # __________________________________________
        if TAGS in inp:
            new = inp[TAGS].create_new_empty()
            new.add_multiple(*inp[TAGS].filter_keys([Tag.TYPES.Link, Tag.TYPES.Node], by='kind'))
            new.add_multiple(*inp[TAGS].filter_keys(((Tag.TYPES.Subcatch, k) for k in inp[SUBCATCHMENTS])))
            inp[TAGS] = new
            # inp[TAGS] = inp[TAGS].slice_section(((Tag.TYPES.Subcatch, k) for k in inp[SUBCATCHMENTS]))

    else:
        for section in [SUBAREAS, INFILTRATION, POLYGONS]:
            if section in inp:
                del inp[section]

        if TAGS in inp:
            inp[TAGS] = inp[TAGS].slice_section([Tag.TYPES.Node, Tag.TYPES.Link], by='kind')

    # __________________________________________
    remove_empty_sections(inp)
    return inp


def create_sub_inp(inp, nodes):
    """
    split model network and only keep nodes.

    Notes:
        CONTROLS not supported

    Args:
        inp (SwmmInput): inp-file data
        nodes (list[str]): list of node labels to keep in inp data

    Returns:
        SwmmInput: filtered inp-file data
    """
    inp = filter_nodes(inp, nodes)
    inp = filter_links_within_nodes(inp, nodes)
    inp = filter_subcatchments(inp, nodes)

    # __________________________________________
    reduce_controls(inp)
    reduce_curves(inp)
    reduce_raingages(inp)
    reduce_pattern(inp)
    remove_empty_sections(inp)
    return inp



import abc
import datetime
import re
import types
from abc import ABC
from collections import UserString
from inspect import isfunction, isclass
import warnings
from numpy import isnan
from pandas import DataFrame, Series
from tqdm.auto import tqdm

from ._type_converter import type2str, is_equal, txt_to_lines
from .section_labels import *
from .section_lists import LINK_SECTIONS, NODE_SECTIONS

SWMM_VERSION = '5.1.015'

COMMENT_STR = ';;'
SEP_INP = COMMENT_STR + "_" * 100
COMMENT_EMPTY_SECTION = COMMENT_STR + ' No Data'


class SwmmInputWarning(UserWarning):
    pass


def head_to_str(head):
    return f'\n{SEP_INP}\n[{head}]\n'


_TYPES_NO_COPY = (int, float, str, datetime.date, datetime.time, UserString)


########################################################################################################################
class CustomDict:
    """
    imitate :class:`collections.UserDict` (:term:`dict-like <mapping>`) but operations only effect self._data
    """
    def __init__(self, d=None, **kwargs):
        if d is None:
            self._data = kwargs
        else:
            if isinstance(d, dict):
                self._data = d
            else:
                self._data = dict(d)

    def __len__(self):
        return self._data.__len__()

    def __getitem__(self, key):
        return self._data.__getitem__(key)

    def __setitem__(self, key, item):
        self._data.__setitem__(key, item)

    def __delitem__(self, key):
        self._data.__delitem__(key)

    def __iter__(self):
        return self._data.__iter__()

    def __contains__(self, key):
        return self._data.__contains__(key)

    def __repr__(self):
        return self._data.__repr__()

    def __str__(self):
        return self._data.__str__()

    def get(self, key, default=None):
        if isinstance(key, list):
            return (self.get(k) for k in key)
        return self._data.get(key) if key in self else default

    def copy(self):
        return type(self)({k: v if isinstance(v, _TYPES_NO_COPY) else v.copy() for k, v in self._data.items()})
        # return type(self)(self._data.copy())

    def values(self):
        return self._data.values()

    def keys(self):
        return self._data.keys()

    def items(self):
        return self._data.items()

    def update(self, d=None, **kwargs):
        self._data.update(d, **kwargs)

    def pop(self, key):
        return self._data.pop(key)

    def __bool__(self):
        return bool(self._data)

    # @property
    # def id(self):
    #     return id(self)


########################################################################################################################
class InpSectionGeneric(CustomDict):
    """
    abstract class for ``.inp``-file sections without objects

    :term:`dict-like <mapping>`"
    """
    _label = ''
    """str: label of the section"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._inp = None

    def __setitem__(self, key, item):
        super().__setitem__(key, item)
        if isinstance(key, str) and ' ' not in key:
            super().__setattr__(key, item)

    def __delitem__(self, key):
        if hasattr(self, key):
            super().__delattr__(key)
        super().__delitem__(key)

    def set_parent_inp(self, inp):
        self._inp = inp

    def get_parent_inp(self):
        return self._inp

    @classmethod
    def from_inp_lines(cls, lines):
        """
        read ``.inp``-file lines and create an section object

        Args:
            lines (str | list[list[str]]): lines in the section of the ``.inp``-file

        Returns:
            InpSectionGeneric: object of the section
        """
        pass

    def to_inp_lines(self, fast=False, sort_objects_alphabetical=False):
        """
        write ``.inp``-file lines of the section object

        Args:
            fast (bool): speeding up conversion

                - :obj:`True`: if no special formation of the input file is needed
                - :obj:`False`: section is converted into a table to prettify string output (slower)

        Returns:
            str: ``.inp``-file lines of the section object
        """
        f = ''
        max_len = len(max(self.keys(), key=len)) + 2
        for sub in self:
            f += '{key}{value}'.format(key=sub.ljust(max_len),
                                       value=type2str(self[sub]) + '\n')
        return f

    @classmethod
    def create_section(cls):
        return cls()


########################################################################################################################
class InpSection(CustomDict):
    """
    class for ``.inp``-file sections with objects (i.e. nodes, links, subcatchments, raingages, ...)
    """

    def __init__(self, section_object):
        """
        create an object for ``.inp``-file sections with objects (i.e. nodes, links, subcatchments, raingages, ...)

        Args:
            section_object (BaseSectionObject-like): object class which is stored in this section.
                This information is used to set the index of the section and
                to decide if the section can be exported (converted to a string) as a table.
        """
        super().__init__()
        self._section_object = section_object
        self._inp = None

    def set_parent_inp(self, inp):
        self._inp = inp

    def get_parent_inp(self):
        return self._inp

    # def __repr__(self):
    #     # return CustomDict.__repr__(self)
    #     return str(self)

    # def __str__(self):
    #     return f'[{self._section_object._section_label}] '  # + ' | '.join((str(s) for s in self.keys()))

    def __setitem__(self, key, value):
        if isinstance(self._identifier, str):
            if not isinstance(key, str):
                warnings.warn(f'Wrong key type (Unknown Behaviour)\n'
                              f'Needed: "{self._identifier}" as string\n'
                              f'Given: {key} of type {type(key)}', SwmmInputWarning)
        else:
            if len(self._identifier) != len(key):
                warnings.warn(f'Wrong number of keys (Unknown Behaviour)\n'
                              f'Needed keys: "{self._identifier}"\n'
                              f'Given keys: "{key}"', SwmmInputWarning)

        if not isinstance(value, self._section_object):
            warnings.warn(f'Wrong section-object type (Unknown Behaviour)\n'
                          f'Needed type: "{self._section_object}"\n'
                          f'Given type: "{type(value)}"', SwmmInputWarning)

        super().__setitem__(key, value)

    @property
    def objects(self):
        """
        all swmm objects in this section

        Returns:
            dict[str, BaseSectionObject]: dictionary of objects with label as key and the object os value
        """
        return self._data

    @property
    def _identifier(self):
        """
        to set the index of the section (key to select an object an index for the dataframe export)

        Returns:
            str | tuple: key of the objects label (can be a single or multiple keys)
        """
        return self._section_object._identifier

    @property
    def _index_labels(self):
        if isinstance(self._identifier, tuple):
            return list(self._identifier)
        else:
            return self._identifier

    @property
    def _label(self):
        """
        get the label of the section

        Returns:
            str: label of the section
        """
        return self._section_object._section_label

    @property
    def _table_inp_export(self):
        # if the section can be exported (converted to a string) as a table.
        return self._section_object._table_inp_export

    def add_multiple(self, *items):
        """
        add objects to section

        Args:
            *items (BaseSectionObject): new objects
        """
        if (len(items) == 1) and isinstance(items[0], types.GeneratorType):
            items = items[0]
        for obj in items:
            self.add_obj(obj)

    def add_obj(self, obj):
        """
        add object to section

        Args:
            obj (BaseSectionObject): new object
        """
        self[obj.get(self._identifier)] = obj

    def add_inp_lines(self, multi_line_args):
        """
        creates and adds objects for each line

        Args:
            multi_line_args (list[list[str]]): lines in the input file section
        """
        self.add_multiple(*self._section_object._convert_lines(multi_line_args))

    @classmethod
    def from_inp_lines(cls, lines, section_class):
        """
        convert the lines of a section to this class and each line to a object

        This function is used for the ``.inp``-file reading

        Args:
            lines (str | list[list[str]]): lines of a section in a ``.inp``-file
            section_class (BaseSectionObject): object class which is stored in this section.

        Returns:
            InpSection: section of the ``.inp``-file
        """
        return section_class.create_section(lines)

    def get_objects(self, sort_alphabetical=False):
        keys = self.keys()
        if sort_alphabetical:
            keys = sorted(keys, key=natural_keys)

        for k in keys:
            yield self[k]

    def to_inp_lines(self, fast=False, sort_objects_alphabetical=False):
        """
        convert the section to a multi-line ``.inp``-file conform string

        This function is used for the ``.inp``-file writing

        Args:
            fast (bool): speeding up conversion

                - :obj:`True`: if no special formation of the input file is needed
                - :obj:`False`: section is converted into a table to prettify string output (slower)

            sort_objects_alphabetical (bool): if objects in a section should be sorted alphabetical |
                default: use order of the read inp-file and append new objects

        Returns:
             str: lines of the ``.inp``-file section
        """
        if not self:  # if empty
            return COMMENT_EMPTY_SECTION

        if fast or not self._table_inp_export:
            return '\n'.join(self.iter_inp_lines(sort_objects_alphabetical))
        else:
            return dataframe_to_inp_string(self.get_dataframe(set_index=True,
                                                              sort_objects_alphabetical=sort_objects_alphabetical))

    def iter_inp_lines(self, sort_objects_alphabetical=False):
        """
        convert the section to a multi-line ``.inp``-file conform string

        This function is used for the ``.inp``-file writing

        Args:
            sort_objects_alphabetical (bool): if objects in a section should be sorted alphabetical |
                default: use order of the read inp-file and append new objects

        Yields:
             str: lines of the ``.inp``-file section
        """
        if self:
            # only show write progress for big files
            n_objects = len(self.keys())
            values = self.get_objects(sort_objects_alphabetical)
            if n_objects > 50_000:
                _iterable = tqdm(values,
                                 desc=self._section_object.__name__,
                                 postfix='Write',
                                 total=n_objects)
            else:
                _iterable = values

            for o in _iterable:
                yield o.to_inp_line()
        else:  # if empty
            yield COMMENT_EMPTY_SECTION

    @property
    def frame(self):
        """convert section to a pandas data-frame

        This property is used for debugging purposes and data analysis of the input data of the swmm model.

        Returns:
            pandas.DataFrame: section as table
        """
        return self.get_dataframe(set_index=True, sort_objects_alphabetical=True)

    def get_dataframe(self, set_index=True, sort_objects_alphabetical=False):
        """convert section to a pandas data-frame

        Args:
            set_index (bool): set object keys as index
            sort_objects_alphabetical (bool): if objects in a section should be sorted alphabetical |
                default: use order of the read inp-file and append new objects

        Returns:
            pandas.DataFrame: section as table
        """
        if not self:  # if empty
            return DataFrame()
        df = DataFrame([i.to_dict_() for i in self.get_objects(sort_objects_alphabetical)])
        if set_index:
            df = df.set_index(self._index_labels)
        return df

    def create_new_empty(self):
        """
        create a new empty section of this kind of section

        Returns:
            InpSection: new empty section
        """
        return type(self)(self._section_object)

    def copy(self):
        """
        get a copy of the section

        Returns:
            InpSection: copy of the section
        """
        new = self.create_new_empty()
        # ΔTime: 18.678 s
        # new._data = deepcopy(self._data)
        # ΔTime: 2.943 s
        # new._data = {k: self[k].copy() for k in self}
        new.add_multiple(v.copy() for v in self.values())
        return new

    def filter_keys(self, keys, by=None):
        """
        filter parts of the section with keys (identifier strings or attribute string)

        Args:
            keys (list | set): list of names to filter by (ether the identifier or the attribute of "by")
            by (str | list[str] | tuple[str]): attribute name of the section object to filter by

        Returns:
            tuple[BaseSectionObject] | list[BaseSectionObject]: filtered objects
        """

        # working with pandas makes it x10 faster
        if by is None:
            filtered_keys = set(self.keys()).intersection(set(keys))
        else:
            f = self.get_dataframe(set_index=False)
            if f.empty:
                return tuple()

            if isinstance(by, (list, set, tuple)):
                f_filtered = f[by].isin(keys).all(axis=1)
                # filtered_keys = (k for k in self if any(map(lambda b: self[k][b] in keys, by)))

            else:
                # filtered_keys = filter(lambda k: self[k][by] in keys, self)
                # filtered_keys = (k for k in self if self[k][by] in keys)
                f_filtered = f[by].isin(keys)

            filtered_keys = f[f_filtered].set_index(self._index_labels).index

        return (self[k] for k in filtered_keys)

    def slice_section(self, keys, by=None):
        """
        filter parts of the section with keys (identifier strings or attribute string)

        Args:
            keys (list | set): list of names to filter by (ether the identifier or the attribute of "by")
            by (str | list[str] | tuple[str]): attribute name of the section object to filter by

        Returns:
            InpSection: new filtered section
        """
        new = self.create_new_empty()
        new.add_multiple(*self.filter_keys(keys, by=by))
        return new


def split_line_with_quotes(line):
    if isinstance(line, (list, tuple)):
        line = ' '.join(line)
    if '"' not in line:
        return line.split()
    return re.findall(r'("[^"]*"|[^" ]+)', line)


########################################################################################################################
class BaseSectionObject(ABC):
    """
    base class for all section objects to unify operations

    sections objects only have __init__ with object parameters

    acts :term:`like a dict <mapping>` (getter and setter)"""
    _identifier = ''
    """str: attribute of an object which will be used as identifiers"""
    _table_inp_export = True
    """bool: if an section is writeable as table. Default ist True"""
    _section_class = InpSection
    """class: section class to identify functionality"""
    _section_label = ''
    """str: label of the section"""

    # @property
    # def section_label(self):
    #     return self._section_label

    __name__ = 'BaseSectionObject'

    @abc.abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    def get(self, key):
        if isinstance(key, (list, tuple, set)):
            return type(key)([self.get(k) for k in key])
        return self.__getattribute__(key)

    def set(self, key, value):
        if not hasattr(self, key):
            raise SwmmInputWarning(f'{key} not a Object attribute | {self}')
        self.__setattr__(key, value)

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, item):
        self.set(key, item)

    def to_dict_(self):
        """
        get all object parameters as dictionary

        Returns:
            dict:
        """
        return vars(self)

    @property
    def attributes(self):
        return tuple(self.to_dict_().keys())

    @property
    def values(self):
        return tuple(self.to_dict_().values())

    @property
    def values_used(self):
        return (v for v in self.values if not (isinstance(v, float) and isnan(v)))

    def __iter__(self):
        for k, v in self.to_dict_().items():
            yield k, v

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self._to_debug_string()

    def __eq__(self, other):
        return isinstance(self, type(other)) and all([is_equal(self[k], other[k]) for k in self.attributes])

    def __hash__(self):
        return tuple([(k, v) for k, v in self]).__hash__()

    # @property
    # def id(self):
    #     return id(self)

    def _to_debug_string(self):
        """for debugging purposes

        string is almost equal to python syntax
        so you could copy it and past it into your code

        Returns:
            str: debug string of the object
        """
        return f'{self.__class__.__name__}({", ".join([f"{k}={repr(v)}" for k, v in self])})'

    def to_inp_line(self):
        """
        convert object to one line of the ``.inp``-file

        for ``.inp``-file writing

        Returns:
            str: SWMM .inp file compatible string
        """
        return ' '.join([type2str(i) for i in self.values_used]).strip()

    @classmethod
    def from_inp_line(cls, *line_args):
        """
        convert line in the ``.inp``-file to the object

        needed if multiple sub-classes of an object are available (i.e. Infiltration)

        Args:
            *line_args (list[str]): arguments in the line

        Returns:
            BaseSectionObject: object of the ``.inp``-file section
        """
        return cls(*line_args)

    def copy(self):
        """
        copy object

        Returns:
            BaseSectionObject or Child: copy of the object
        """
        # return type(self)(*self.values_used)
        return type(self)(**self.to_dict_())

    @classmethod
    def create_section(cls, lines=None):
        """
        creates a new section for the ``.inp``-file of this object and ads objects described in `lines`

        Args:
            lines:

        Returns:
            InpSection: new section of this object
        """
        sec = cls._section_class(cls)
        # import sys
        if lines is not None:
            # print(f'BYTES: {len(lines):>9_d}', )
            if isinstance(lines, str):
                if len(lines) > 10_000_000:
                    n_lines = lines.strip().count('\n') + 1
                    # to create a progressbar in the reading process
                    # only needed with big (> 200 MB) files
                    lines_iter = tqdm(txt_to_lines(lines), desc=cls.__name__, total=n_lines, postfix='Read')
                else:
                    lines_iter = txt_to_lines(lines)
            else:
                lines_iter = lines

            sec.add_inp_lines(lines_iter)

            if isinstance(lines_iter, tqdm):
                lines_iter.close()

        return sec

    @classmethod
    def from_inp_lines(cls, lines):
        return cls.create_section(lines)

    @classmethod
    def _convert_lines(cls, multi_line_args):
        """
        convert the ``.inp``-file section

        creates and yields an object for each line

        Args:
            multi_line_args (list[list[str]]): lines in the input file section

        Yields:
            BaseSectionObject: object of the ``.inp``-file section
        """
        # overwrite if each object has multiple lines
        for line_args in multi_line_args:
            yield cls.from_inp_line(*line_args)


########################################################################################################################
def dataframe_to_inp_string(df):
    """convert a data-frame into a multi-line string

    used to make a better readable .inp file and for debugging

    Args:
        df (pandas.DataFrame): section table

    Returns:
        str: .inp file conform string for one section
    """
    if df.empty:
        return COMMENT_EMPTY_SECTION

    if isinstance(df, Series):
        return df.apply(type2str).to_string()

    c = df.copy()
    if c.columns.name is None:
        c.columns.name = COMMENT_STR
    else:
        if not c.columns.name.startswith(COMMENT_STR):
            c.columns.name = COMMENT_STR + c.columns.name

    if c.index.name is not None:
        if not c.index.name.startswith(COMMENT_STR):
            c.index.name = COMMENT_STR + c.index.name

    if c.index._typ == 'multiindex':
        if c.index.names is not None:
            if not c.index.levels[0].name.startswith(COMMENT_STR):
                c.index.set_names(';' + c.index.names[0], level=0, inplace=True)
                # because pandas 1.0
                # c.index.levels[0].name = ';' + c.index.levels[0].name

    return c.applymap(type2str).to_string(sparsify=False,
                                          line_width=999999,
                                          max_rows=999999,
                                          max_cols=999999,
                                          max_colwidth=999999)


########################################################################################################################
def convert_section(head, lines, converter):
    """
    convert section string to a section object

    Args:
        head (str): header of the section
        lines (str): lines in the section
        converter (dict): dict of converters assigned to header {header: converter]

    Returns:
        str | InpSection | InpSectionGeneric: converted section
    """
    if head in converter:
        section_ = converter[head]

        if isfunction(section_):  # section_ ... converter function
            return section_(lines)

        elif isclass(section_):  # section_ ... type/class
            return section_.from_inp_lines(lines)

        else:
            warnings.warn(f'Type of converter ({type(section_)}) for Section "{head}" not implemented. Section will not be converted.', SwmmInputWarning)
    else:
        warnings.warn(f'Section "{head}" not implemented. Section will not be converted.', SwmmInputWarning)

    return lines.replace(SEP_INP, '').strip()


########################################################################################################################
SECTIONS_ORDER_MP = ([
                         TITLE,
                         OPTIONS,
                         REPORT,
                         EVAPORATION,
                         TEMPERATURE

                     ] + NODE_SECTIONS +
                     [
                         DWF,
                         INFLOWS

                     ] + LINK_SECTIONS +
                     [

                         LOSSES,
                         XSECTIONS,
                         TRANSECTS,

                         CURVES,
                         TIMESERIES,
                         RAINGAGES,
                         PATTERNS,

                         SUBCATCHMENTS,
                         SUBAREAS,
                         INFILTRATION,

                         POLLUTANTS,
                         LOADINGS,
                     ])

SECTION_ORDER_DEFAULT = [TITLE,
                         OPTIONS,
                         FILES,
                         EVAPORATION,
                         TEMPERATURE,
                         RAINGAGES,
                         SUBCATCHMENTS,
                         SUBAREAS,
                         INFILTRATION,
                         LID_CONTROLS,
                         LID_USAGE,
                         AQUIFERS,
                         GROUNDWATER,
                         GWF,
                         SNOWPACKS,
                         JUNCTIONS,
                         OUTFALLS,
                         STORAGE,
                         CONDUITS,
                         PUMPS,
                         ORIFICES,
                         WEIRS,
                         OUTLETS,
                         XSECTIONS,
                         TRANSECTS,
                         LOSSES,
                         CONTROLS,
                         POLLUTANTS,
                         LANDUSES,
                         COVERAGES,
                         LOADINGS,
                         BUILDUP,
                         WASHOFF,
                         TREATMENT,
                         INFLOWS,
                         DWF,
                         HYDROGRAPHS,
                         RDII,
                         CURVES,
                         TIMESERIES,
                         PATTERNS,
                         REPORT,
                         ADJUSTMENTS,
                         TAGS,
                         MAP,
                         COORDINATES,
                         VERTICES,
                         POLYGONS,
                         SYMBOLS,
                         LABELS,
                         BACKDROP,
                         PROFILES]


def check_order(inp, order_list=None):
    if order_list is None:
        order_list = SECTION_ORDER_DEFAULT
    order = [order_list.index(o) if o in order_list else len(order_list) for o in inp]
    return all((order[i+1] - order[i]) > 0 for i in range(len(order)-1))


def _sort_by(key, sections_order):
    if key in sections_order:
        return sections_order.index(key)
    else:
        return len(sections_order)


re_int = re.compile('(\d+)')


def natural_keys(text):
    if isinstance(text, str):
        return [int(text) if text.isdigit() else text for text in re_int.split(text)]
    else:
        return [*(natural_keys(t) for t in text)]


def section_to_string(section, fast=True, sort_objects_alphabetical=False):
    """
    create a string of a section in an ``.inp``-file

    Args:
        section (InpSection | InpSectionGeneric):
            section of an ``.inp``-file
        fast (bool): don't use any formatting else format as table
        sort_objects_alphabetical (bool): if objects in a section should be sorted alphabetical |
            default: use order of the read inp-file and append new objects

    Returns:
        str: string of the ``.inp``-file section
    """
    if isinstance(section, str):  # Title
        return section.replace(SEP_INP, '').strip()

    # ----------------------
    elif isinstance(section, list):  # V1
        return '\n'.join([type2str(line) for line in section])

    # ----------------------
    elif isinstance(section, dict):  # V2
        max_len = len(max(section.keys(), key=len)) + 2
        f = ''
        for sub in section:
            f += '{key}{value}'.format(key=sub.ljust(max_len), value=type2str(section[sub]) + '\n')
        return f

    # ----------------------
    elif isinstance(section, (DataFrame, Series)):  # V3
        return dataframe_to_inp_string(section)

    # ----------------------
    elif isinstance(section, InpSectionGeneric):  # V4
        return section.to_inp_lines(fast=fast)

    elif isinstance(section, InpSection):  # V4
        return section.to_inp_lines(fast=fast, sort_objects_alphabetical=sort_objects_alphabetical)

    elif section is None:
        return ''

    else:
        print(section, '?'*10)


def iter_section_lines(section, sort_objects_alphabetical=False):
    if isinstance(section, str):  # Title
        yield section.replace(SEP_INP, '').strip()

    # ----------------------
    elif isinstance(section, list):  # V1
        for line in section:
            yield type2str(line)

    # ----------------------
    elif isinstance(section, dict):  # V2
        max_len = len(max(section.keys(), key=len)) + 2
        for sub in section:
            yield'{key}{value}'.format(key=sub.ljust(max_len), value=type2str(section[sub]) + '\n')

    # ----------------------
    elif isinstance(section, (DataFrame, Series)):  # V3
        yield dataframe_to_inp_string(section)

    # ----------------------
    elif isinstance(section, InpSectionGeneric):
        yield section.to_inp_lines(fast=True)
    elif isinstance(section, InpSection):  # V4
        for line in section.iter_inp_lines(sort_objects_alphabetical=sort_objects_alphabetical):
            yield line
